import numpy as np
from bamboo.utils import Map, ProcessMap, ThreadMap
from bamboo.writer import (
    SharedMemoryWriter,
    MemmapWriter,
    NumpyWriter,
    ZarrWriter,
    inject_writer,
)


def test_writer_shared_process():
    writer = SharedMemoryWriter((8, 8), np.uint8)
    writer.create()
    indices = range(8)

    @inject_writer(writer)
    def _worker(output, i):
        output[i, i] = i

    _ = ProcessMap(_worker, indices, n_jobs=2, silent=True)

    data = writer.output.copy()
    writer.cleanup()
    assert np.all(np.diag(data) == np.arange(8))


def test_writer_shared_threads():
    writer = SharedMemoryWriter((8, 8), np.uint8)
    writer.create()
    indices = range(8)

    @inject_writer(writer)
    def _worker(output, i):
        output[i, i] = i

    _ = ThreadMap(_worker, indices, n_jobs=2, silent=True)

    data = writer.output.copy()
    writer.cleanup()
    assert np.all(np.diag(data) == np.arange(8))


def test_writer_memmap_process():
    writer = MemmapWriter("memmap_a.npy", (8, 8), np.uint8)
    writer.create()
    indices = range(8)

    @inject_writer(writer)
    def _worker(output, i):
        output[i, i] = i

    _ = ProcessMap(_worker, indices, n_jobs=2, silent=True)

    data = writer.output.copy()
    writer.cleanup()
    assert np.all(np.diag(data) == np.arange(8))


def test_writer_memmap_threads():
    writer = MemmapWriter("memmap_b.npy", (8, 8), np.uint8)
    writer.create()
    indices = range(8)

    @inject_writer(writer)
    def _worker(output, i):
        output[i, i] = i

    _ = ThreadMap(_worker, indices, n_jobs=2, silent=True)

    data = writer.output.copy()
    writer.cleanup()
    assert np.all(np.diag(data) == np.arange(8))


def test_writer_zarr_process():
    writer = ZarrWriter("array_a.zarr", (2, 2), (8, 8), np.uint8)
    writer.create()
    indices = np.arange(8)

    @inject_writer(writer)
    def _worker(output, i):
        output[i, i] = i

    _ = ProcessMap(_worker, np.split(indices, 8 // 2), n_jobs=2, silent=True)

    data = writer.output[:]
    writer.cleanup()
    assert np.all(np.diag(data) == np.arange(8))


def test_writer_zarr_threads():
    writer = ZarrWriter("array_b.zarr", (1, 1), (8, 8), np.uint8)
    writer.create()
    indices = np.arange(8)

    @inject_writer(writer)
    def _worker(output, i):
        output[i, i] = i

    _ = ThreadMap(_worker, np.split(indices, 8 // 2), n_jobs=2, silent=True)

    data = writer.output[:]
    writer.cleanup()
    assert np.all(np.diag(data) == np.arange(8))


def test_writer_numpy_map():
    writer = NumpyWriter((8, 8), np.uint8)
    writer.create()
    indices = range(8)

    @inject_writer(writer)
    def _worker(output, i):
        output[i, i] = i

    _ = Map(_worker, indices, silent=True)

    data = writer.output.copy()
    writer.cleanup()
    assert np.all(np.diag(data) == np.arange(8))



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

from concurrent.futures import ThreadPoolExecutor
from joblib import Parallel, delayed, parallel_backend
from tqdm import tqdm
from typing import Callable, Iterable, Any


class Map:
    """Apply a function to each element in an iterable.

    Parameters
    ----------
    func : Callable
        An arbitrary function valid for the provided iterable.
    iterable : Iterable
        An iterable object.
    silent : bool
        If True, suppress the progress bar.
    desc : None | str
        Optional user-provided progress bar description.

    Returns
    -------
    Any
        `func` output.
    """

    def __new__(
        cls,
        func: Callable,
        iterable: Iterable,
        silent: bool = False,
        desc: None | str = None,
    ) -> Any:
        tqdm._instances.clear()
        return [func(i) for i in tqdm(iterable, disable=silent, desc=desc)]


class ProcessMap:
    """Apply a function to each element in an iterable in parallel.

    Parameters
    ----------
    func : Callable
        An arbitrary function valid for the provided iterable.
    iterable : Iterable
        An iterable object.
    n_jobs : None | int
        Number of parallel jobs.
    silent : bool
        If True, suppress the progress bar.
    desc : None | str
        Optional user-provided progress bar description.
    kwargs : Any
        Additional arguments to joblib.Parallel.

    Returns
    -------
    Any
        `func` output.
    """

    def __new__(
        cls,
        func: Callable,
        iterable: Iterable,
        n_jobs: None | int = None,
        silent: bool = False,
        desc: None | str = None,
        **kwargs,
    ) -> Any:
        tqdm._instances.clear()
        with parallel_backend("loky"):
            return Parallel(n_jobs=n_jobs, prefer="processes", **kwargs)(
                delayed(func)(i) for i in tqdm(iterable, disable=silent, desc=desc)
            )


class ThreadMap:
    """Apply a function to each element in an iterable using multiple threads.

    Parameters
    ----------
    func : Callable
        An arbitrary function valid for the provided iterable.
    iterable : Iterable
        An iterable object.
    n_jobs : None | int
        Number of parallel jobs.
    silent : bool
        If True, suppress the progress bar.
    desc : None | str
        Optional user-provided progress bar description.
    kwargs : Any
        Additional arguments to ThreadPoolExecutor.

    Returns
    -------
    Any
        `func` output.
    """

    def __new__(
        cls,
        func: Callable,
        iterable: Iterable,
        n_jobs: None | int = None,
        silent: bool = False,
        desc: None | str = None,
        **kwargs,
    ) -> Any:
        tqdm._instances.clear()
        with ThreadPoolExecutor(max_workers=n_jobs, **kwargs) as executor:
            return list(
                tqdm(executor.map(func, iterable), total=len(iterable), desc=desc)
            )


def apply_map(
    func: Callable,
    iterable: Iterable,
    n_jobs: None | int = None,
    prefer: None | str = None,
    silent: bool = False,
    desc: None | str = None,
    **kwargs,
) -> Any:
    """Selects and runs the appropriate map given parallelism settings.

    Parameters
    ----------
    func : Callable
        An arbitrary function valid for the provided iterable.
    iterable : Iterable
        An iterable object.
    n_jobs : None | int
        Number of parallel jobs.
    prefer : None | str
        Either None, 'processes, or 'threads'. If n_jobs and prefer are both
        None, then no parallelism will be used (n_jobs == 1). If n_jobs > 1 and
        prefer is None, then 'processes' will be used. If prefer is 'processes'
        and n_jobs is None, all available processes will be used.
    silent : bool
        If True, suppress the progress bar.
    desc : None | str
        Optional user-provided progress bar description.
    kwargs : Any
        Additional arguments to ThreadPoolExecutor.

    Returns
    -------
    Any
        Iteration output.
    """
    if isinstance(n_jobs, int):
        if n_jobs < 1:
            raise ValueError("'n_jobs' must be greater than 0.")
    else:
        if n_jobs is not None:
            raise TypeError("'n_jobs' must be None or an `int`.")

    if prefer not in [None, "threads", "processes"]:
        raise ValueError("'prefer' must be one of: None, 'threads', or 'processes'.")

    if (n_jobs is None and prefer is None) or n_jobs == 1:
        return Map(func, iterable, silent=silent, desc=desc)

    if prefer is None or prefer == "processes":
        return ProcessMap(
            func, iterable, n_jobs=n_jobs, silent=silent, desc=desc, **kwargs
        )

    return ThreadMap(func, iterable, n_jobs=n_jobs, silent=silent, desc=desc)



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

from anci import main

from .brightfield import brightfield

__all__ = ["brightfield"]

if __name__ == "__main__":
    main("bamboo: flexible gigapixel image analysis for the spatial biology era")



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

import matplotlib.pyplot as plt
from typing import Any, Iterator
from bamboo.tile import Tile


class Tiles(Iterator):
    """An iterator that yields image tiles, optionally with coordinates or indices.

    Parameters
    ----------
    image : Any
        An Image object with `tile` and `resize` methods.
    tile_w : int
        Tile width in pixels.
    tile_h : int
        Tile height in pixels.
    mpp : float | None
        Target microns per pixel for resizing.
    resample : str | None
        Resampling filter if resizing is performed.
    overlap_x, overlap_y : int
        Tile overlap in pixels.
    mode : str
        Determines what is returned during iteration:
        - "default": tile only
        - "coords": (coords, tile)
        - "enumerate": (index, tile)
        - "enumerate+coords": (index, coords, tile)
    _scale : float | None
        Optional scale override (for testing/debugging).
    """

    def __init__(
        self,
        image: Any,
        tile_w: int,
        tile_h: int,
        mpp: None | float = None,
        resample: str | None = "bilinear",
        overlap_x: int = 0,
        overlap_y: int = 0,
        mode: str = "default",
        _scale: None | float = None,
    ):
        self.image = image
        self.mode = mode.lower()
        if self.mode not in {"default", "coords", "enumerate", "enumerate+coords"}:
            raise ValueError(
                f"Invalid mode '{self.mode}'. Must be one of "
                "'default', 'coords', 'enumerate', 'enumerate+coords'."
            )

        # Handle scaling and mpp
        if _scale:
            self.scale = _scale
        else:
            self.scale = None
            if mpp is not None:
                if not hasattr(image, "mpp"):
                    raise ValueError("Image must have an 'mpp' attribute to use 'mpp'.")
                self.scale = mpp / image.mpp

        self.image_w = image.w
        self.image_h = image.h

        self.crop_w = int(round(tile_w * (self.scale if self.scale else 1.0)))
        self.crop_h = int(round(tile_h * (self.scale if self.scale else 1.0)))

        if self.crop_w <= 0 or self.crop_h <= 0:
            raise ValueError("Invalid tile size after rescaling.")

        self.overlap_x = overlap_x
        self.overlap_y = overlap_y
        self.tile_w = tile_w
        self.tile_h = tile_h
        self.resize = bool(self.scale)
        self.resample = resample

        # Compute coordinates
        x_min, y_min = 0, 0
        x_max, y_max = image.w, image.h
        step_x = self.crop_w - self.overlap_x
        step_y = self.crop_h - self.overlap_y

        if step_x <= 0 or step_y <= 0:
            raise ValueError("Overlap must be less than tile dimension.")

        self.coordinates = []
        for y in range(y_min, y_max, step_y):
            for x in range(x_min, x_max, step_x):
                orig_x, orig_y = x, y
                if x + self.crop_w > image.w:
                    x = image.w - self.crop_w
                if y + self.crop_h > image.h:
                    y = image.h - self.crop_h
                overlap_w = max(0, orig_x - x)
                overlap_h = max(0, orig_y - y)
                self.coordinates.append(
                    (x, y, self.crop_w, self.crop_h, overlap_w, overlap_h)
                )

        self.n_tiles = len(self.coordinates)
        self._index = 0
        self._coordinates = iter(self.coordinates)

    @property
    def hw(self):
        return (self.tile_h, self.tile_w) if self.resize else (self.crop_h, self.crop_w)

    def __len__(self):
        return len(self.coordinates)

    def __iter__(self):
        self._index = 0
        self._coordinates = iter(self.coordinates)
        return self

    def __next__(self):
        if self._index >= self.n_tiles:
            raise StopIteration

        coord = next(self._coordinates)
        x, y, w, h, overlap_w, overlap_h = coord
        tile = self.image.tile(x, y, w, h)
        if self.resize:
            tile = tile.resize(self.tile_w, self.tile_h, self.resample)

        idx = self._index
        self._index += 1

        if self.mode == "default":
            return tile
        elif self.mode == "coords":
            return coord, tile
        elif self.mode == "enumerate":
            return idx, tile
        elif self.mode == "enumerate+coords":
            return idx, coord, tile

    def exclude(self, coords: list[tuple[int, int, int, int]]) -> None:
        """Remove tiles that overlap with any in the given exclusion coordinates.

        Parameters
        ----------
        coords : list of (x, y, w, h)
            Tile coordinates to exclude (e.g., from QC results).
        """

        def overlaps(a, b):
            ax, ay, aw, ah, *_ = a
            bx, by, bw, bh, *_ = b
            return ax < bx + bw and ax + aw > bx and ay < by + bh and ay + ah > by

        # Filter out any coordinate that overlaps with an exclusion region
        filtered = []
        for coord in self.coordinates:
            if any(overlaps(coord, ex) for ex in coords):
                continue
            filtered.append(coord)

        # Update iterator internals
        self.coordinates = filtered
        self.n_tiles = len(filtered)
        self._index = 0
        self._coordinates = iter(self.coordinates)

    def show(
        self,
        ax: None | plt.Axes = None,
        figsize: None | tuple[int, int] = None,
        edgecolor: str | None = "black",
        facecolor: str | None = "green",
        linewidth: int | float = 1,
        alpha: int | float = 0.1,
        t: int = 5000,
        force: bool = False,
        height: int = None,
        width: int = None,
        resample: str | None = "bilinear",
        hide_axes: bool = True,
    ) -> plt.Axes:
        """Visualize tile overlaid on image thumbnail.

        Parameters
        ----------
        ax : None | plt.Axes
            Optional matplot axes object.
        figsize : None | tuple[int, int]
            Option figure size.
        edgecolor : str | None
            Edge color around plotted tiles.
        facecolor : str | None
            Face color of plotted tiles.
        linewidth : int | float
            Edge line width around plotted tiles.
        t: int
            Find thumbnail with height + width closest to this value.
        force: bool
            If True and image is not pyramidal, then the thumbnail will be
            generated at the height and width of the full resolution image.
        height: None | int
            Optionally resize the image thumbnail to a user-provided height.
        width: None | int
            Optionally resize the image thumbnail to a user-provided width.
        resample: str | None
            Resampling filter ('bilinear', 'bicubic', 'lanczos', 'nearest').
        alpha : int | float
            Alpha of tile colors.
        hide_axes : bool
            If True, then all x-axis, y-axis, and spines are removed.

        Returns
        -------
        plt.Axes
            Matplotlib axes object of tile.
        """
        thumbnail = self.image.thumbnail(
            t=t, force=force, height=height, width=width, resample=resample
        ).to_numpy()

        ih, iw, _ = self.image.shape
        th, tw, _ = thumbnail.shape
        scale = th / ih

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        ax.imshow(thumbnail)
        for coord in self.coordinates:
            x, y, w, h, *_ = coord
            x = int(x * scale)
            y = int(y * scale)
            w = int(w * scale)
            h = int(h * scale)
            rect = plt.Rectangle(
                (x, y),
                w,
                h,
                edgecolor=edgecolor,
                facecolor=facecolor,
                linewidth=linewidth,
                alpha=alpha,
            )
            ax.add_patch(rect)

        if hide_axes:
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)
            for pos in ["left", "right", "top", "bottom"]:
                ax.spines[pos].set_visible(False)

        return ax



import pytest
import numpy as np
import openslide
from pathlib import Path
from zarr.core.group import Group as ZarrGroup
from zarr.core.group import Array as ZarrArray

from bamboo.backend._brightfield import (
    BrightfieldTiffBackend,
    BrightfieldOpenslideBackend,
    BrightfieldMemoryBackend,
)

from tests._fixtures.slides import (
    SlideBif,
    SlideNdpi,
    SlidePhilips,
    SlideLeica,
    SlideSvs,
    SlideTiff,
    SlideTrestle,
    SlideOmeTiff,
    SlideError,
)

from tests._fixtures.images import (
    RgbBmp,
    RgbJpeg,
    RgbNpy,
    RgbPbm,
    RgbPng,
    RgbTga,
    RgbWebp,
)

FORMATS_TIFF_BACKEND = [
    SlideNdpi,
    SlidePhilips,
    SlideLeica,
    SlideSvs,
    SlideTiff,
    SlideOmeTiff,
]

FORMATS_OPENSLIDE_BACKEND = [
    SlideBif,
    SlideNdpi,
    SlidePhilips,
    SlideSvs,
    SlideTiff,
    SlideTrestle,
]

FORMATS_MEMORY_BACKEND = [
    RgbBmp,
    RgbJpeg,
    RgbNpy,
    RgbPbm,
    RgbPng,
    RgbTga,
    RgbWebp,
]


success_tiff = SlideSvs()
failure_tiff = SlideError()


def test_tiff_init_str():
    backend = BrightfieldTiffBackend(success_tiff.path)
    for attr in [
        "path",
        "backend",
        "pyramidal",
        "levels",
        "levels_keys",
        "levels_shape",
        "l",
        "h",
        "w",
        "c",
        "mpp",
        "mpp_x",
        "mpp_y",
    ]:
        assert hasattr(backend, attr)


def test_tiff_init_pathlib():
    backend = BrightfieldTiffBackend(Path(success_tiff.path))
    for attr in [
        "path",
        "backend",
        "pyramidal",
        "levels",
        "levels_keys",
        "levels_shape",
        "l",
        "h",
        "w",
        "c",
        "mpp",
        "mpp_x",
        "mpp_y",
    ]:
        assert hasattr(backend, attr)


def test_tiff_validate_failure():
    with pytest.raises(ValueError):
        _ = BrightfieldTiffBackend(failure_tiff.path)


def test_tiff_read():
    backend = BrightfieldTiffBackend(success_tiff.path)
    assert hasattr(backend, "data")
    assert isinstance(backend.data, ZarrGroup)


def test_tiff_formats():
    for fmt in FORMATS_TIFF_BACKEND:
        path = Path(fmt.path)
        backend = BrightfieldTiffBackend(path)
        assert hasattr(backend, "data")
        assert isinstance(backend.data, (ZarrGroup, ZarrArray))
        assert backend.mpp_x == fmt.mpp_x
        assert backend.mpp_y == fmt.mpp_y
        if fmt.mpp_x is not None and fmt.mpp_y is not None:
            assert backend.mpp == (fmt.mpp_x + fmt.mpp_y) / 2.0
        if backend.levels_keys:
            assert backend.pyramidal
            assert backend.l == "0"
        else:
            assert not backend.pyramidal
            assert backend.l is None
        assert backend.w == fmt.width
        assert backend.h == fmt.height
        assert backend.c == fmt.channels
        assert backend[:100, :100].shape == (100, 100, fmt.channels)


def test_tiff_indexing():
    backend = BrightfieldTiffBackend(success_tiff.path, level="2")
    h, w, c = backend.h, backend.w, 3
    # Simplified check of shapes
    assert backend[0:5, 0:5, 0].shape == (5, 5)
    assert backend[0:2, 0:2, 0:c].shape == (2, 2, c)
    assert backend[...].shape == (h, w, c)
    assert backend[..., 0].shape == (h, w)
    assert backend[0, ..., 0].shape == (w,)
    assert backend[0].shape == (w, c)
    assert backend[0:10].shape[0] == 10


success_os = SlideSvs()
failure_os = SlideError()


def test_openslide_init_str():
    backend = BrightfieldOpenslideBackend(success_os.path)
    for attr in [
        "path",
        "backend",
        "pyramidal",
        "levels",
        "levels_keys",
        "levels_shape",
        "l",
        "h",
        "w",
        "c",
        "mpp",
        "mpp_x",
        "mpp_y",
    ]:
        assert hasattr(backend, attr)


def test_openslide_init_pathlib():
    backend = BrightfieldOpenslideBackend(Path(success_os.path))
    for attr in [
        "path",
        "backend",
        "pyramidal",
        "levels",
        "levels_keys",
        "levels_shape",
        "l",
        "h",
        "w",
        "c",
        "mpp",
        "mpp_x",
        "mpp_y",
    ]:
        assert hasattr(backend, attr)


def test_openslide_validate_failure():
    with pytest.raises(ValueError):
        _ = BrightfieldOpenslideBackend(failure_os.path)


def test_openslide_read():
    backend = BrightfieldOpenslideBackend(success_os.path)
    assert hasattr(backend, "data")
    assert isinstance(backend.data, openslide.OpenSlide)


def test_openslide_formats():
    for fmt in FORMATS_OPENSLIDE_BACKEND:
        path = Path(fmt.path)
        backend = BrightfieldOpenslideBackend(path)
        assert backend.mpp_x == fmt.mpp_x
        assert backend.mpp_y == fmt.mpp_y
        if fmt.mpp_x and fmt.mpp_y:
            assert backend.mpp == (fmt.mpp_x + fmt.mpp_y) / 2.0
        if backend.levels_keys:
            assert backend.pyramidal
            assert backend.l == 0
        else:
            assert not backend.pyramidal
            assert backend.l is None
        assert backend.w == fmt.width
        assert backend.h == fmt.height
        assert backend.c == fmt.channels
        assert backend[:100, :100].shape == (100, 100, fmt.channels)


def test_openslide_indexing():
    backend = BrightfieldOpenslideBackend(success_os.path, level=2)
    h, w, c = backend.h, backend.w, 3
    assert backend[0:5, 0:5, 0].shape == (5, 5)
    assert backend[0:2, 0:2, 0:c].shape == (2, 2, c)
    assert backend[...].shape == (h, w, c)
    assert backend[..., 0].shape == (h, w)
    assert backend[0, ..., 0].shape == (w,)
    assert backend[0].shape == (w, c)
    assert backend[0:10].shape[0] == 10


success_mem = RgbBmp()
failure_mem = SlideError()


def test_memory_init_str():
    backend = BrightfieldMemoryBackend(success_mem.path)
    for attr in [
        "path",
        "backend",
        "pyramidal",
        "levels",
        "levels_keys",
        "levels_shape",
        "l",
        "h",
        "w",
        "c",
        "mpp",
        "mpp_x",
        "mpp_y",
    ]:
        assert hasattr(backend, attr)


def test_memory_init_pathlib():
    backend = BrightfieldMemoryBackend(Path(success_mem.path))
    for attr in [
        "path",
        "backend",
        "pyramidal",
        "levels",
        "levels_keys",
        "levels_shape",
        "l",
        "h",
        "w",
        "c",
        "mpp",
        "mpp_x",
        "mpp_y",
    ]:
        assert hasattr(backend, attr)


def test_memory_validate_failure():
    with pytest.raises(ValueError):
        _ = BrightfieldMemoryBackend(failure_mem.path)


def test_memory_validate_level():
    with pytest.raises(ValueError):
        _ = BrightfieldMemoryBackend(success_mem.path, level=2)


def test_memory_read():
    backend = BrightfieldMemoryBackend(success_mem.path)
    assert hasattr(backend, "data")
    assert isinstance(backend.data, np.ndarray)


def test_memory_formats():
    for fmt in FORMATS_MEMORY_BACKEND:
        path = Path(fmt.path)
        backend = BrightfieldMemoryBackend(path)
        assert backend.mpp_x == fmt.mpp_x
        assert backend.mpp_y == fmt.mpp_y
        if fmt.mpp_x and fmt.mpp_y:
            assert backend.mpp == (fmt.mpp_x + fmt.mpp_y) / 2.0
        assert not backend.pyramidal
        assert backend.l is None
        assert backend.w == fmt.width
        assert backend.h == fmt.height
        assert backend.c == fmt.channels
        assert backend[:100, :100].shape == (100, 100, fmt.channels)


def test_memory_indexing():
    backend = BrightfieldMemoryBackend(success_mem.path)
    h, w, c = backend.h, backend.w, 3
    assert backend[0:5, 0:5, 0].shape == (5, 5)
    assert backend[0:2, 0:2, 0:c].shape == (2, 2, c)
    assert backend[...].shape == (h, w, c)
    assert backend[..., 0].shape == (h, w)
    assert backend[0, ..., 0].shape == (w,)
    assert backend[0].shape == (w, c)
    assert backend[0:10].shape[0] == 10



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

import tifffile
import importlib.util
import numpy as np
import re
import xml.etree.ElementTree as ET
import zarr

from abc import ABC, abstractmethod
from bamboo.types import IndexKey, Numeric
from dataclasses import dataclass, field
from numpy.typing import NDArray
from pathlib import Path
from PIL import Image
from tifffile import TiffFile
from typing import Self, Final, Set


@dataclass(frozen=True)
class BrightfieldBackendConfig:
    name: str
    formats: Set[str] = field(default_factory=set)

    def supports_format(self, path: Path) -> bool:
        return path.suffix.lower() in self.formats


BRIGHTFIELD_MEMORY_CONFIG = BrightfieldBackendConfig(
    name="memory",
    formats={
        ".apng",
        ".blp",
        ".bmp",
        ".bufr",
        ".bw",
        ".cur",
        ".dcx",
        ".dds",
        ".dib",
        ".emf",
        ".eps",
        ".fit",
        ".fits",
        ".flc",
        ".fli",
        ".ftc",
        ".ftu",
        ".gbr",
        ".gif",
        ".grib",
        ".h5",
        ".hdf",
        ".icb",
        ".icns",
        ".ico",
        ".iim",
        ".im",
        ".j2c",
        ".j2k",
        ".jfif",
        ".jp2",
        ".jpc",
        ".jpe",
        ".jpeg",
        ".jpf",
        ".jpg",
        ".jpx",
        ".mpeg",
        ".mpg",
        ".msp",
        ".pbm",
        ".pcd",
        ".pcx",
        ".pfm",
        ".pgm",
        ".png",
        ".pnm",
        ".ppm",
        ".ps",
        ".psd",
        ".pxr",
        ".qoi",
        ".ras",
        ".rgb",
        ".rgba",
        ".sgi",
        ".tga",
        ".tif",
        ".tiff",
        ".vda",
        ".vst",
        ".webp",
        ".wmf",
        ".xbm",
        ".xpm",
        ".npy",
    },
)

BRIGHTFIELD_TIFF_CONFIG = BrightfieldBackendConfig(
    name="tifffile", formats={".ndpi", ".scn", ".svs", ".tif", ".tiff", ".ome.tiff"}
)

BRIGHTFIELD_OPENSLIDE_CONFIG = BrightfieldBackendConfig(
    name="openslide",
    formats={
        ".bif",
        ".czi",
        ".mrxs",
        ".ndpi",
        ".scn",
        ".svs",
        ".svslide",
        ".tif",
        ".tiff",
        ".vms",
        ".vmu",
    },
)


class BrightfieldBackend(ABC):
    """Base class for various histology image backends.

    Parameters
    ----------
    path : str or Path
        Path to the input image file.
    mpp : None | float
        User-provided microns per pixel.
    level : None | int | str
        User-provided level to read.

    Raises
    ------
    ValueError
        If `path` is not a string or pathlib.Path.
    FileNotFoundError
        If the specified path does not exist.
    OSError
        If the path is a directory instead of a file.

    Attributes
    ----------
    path : Path
        Path to the input image file.
    backend : str
        Name of the backend used for reading.
    pyramidal : bool | None
        Whether the image is in pyramidal format.
    levels : int | None
        Number of image pyramid levels.
    levels_keys : list[str | int]
        If pyramidal, key/integer for indexing a specific level.
    levels_shape : dict[str, tuple[int, int, int]]
        If pyramidal, height, width, and channels for each level.
    mpp : float | None
        Micron per pixel information.
    mpp_x : float | None
        Micron per pixel information across x-axis.
    mpp_x : float | None
        Micron per pixel information across x-orientation.
    l : int | None
        Current level.
    h : int | None
        Image height at the current level.
    w : int | None
        Image width at the current level.
    c : int | None
        Number of channels in the image.
    """

    formats: Final[list[str]] = []
    """Backend-specific supported image formats."""

    def __init__(
        self, path: str | Path, mpp: None | float = None, level: None | int | str = None
    ) -> Self:
        if not isinstance(path, (str, Path)):
            raise ValueError("'path' must be a str or pathlib.Path.")

        if not (path := Path(path)).exists():
            raise FileNotFoundError("'path' does not exist.")

        if path.is_dir():
            raise OSError("'path' must be a file not a directory.")

        self.path = path
        self.backend: str = None
        self.pyramidal: bool | None = None
        self.dtype: Numeric | None = None

        self.levels: int | None = None
        self.levels_keys: list[str | int] = None
        self.levels_shape: dict[str | int, tuple[int, int, int]] = None

        self.mpp_x: float | None = mpp
        self.mpp_y: float | None = mpp

        self.l: str | None = level
        self.h: int | None = None
        self.w: int | None = None
        self.c: int | None = None

        self._validate()
        if mpp is None:
            self._resolution()

        self._read()

    @property
    def mpp(self) -> float | None:
        """Set average microns per pixel."""
        if self.mpp_x is not None and self.mpp_y is not None:
            return (self.mpp_x + self.mpp_y) / 2.0
        return None

    @abstractmethod
    def _resolution(self):
        """Extract image resolution metadata."""
        pass

    @abstractmethod
    def _validate(self):
        """Validate input image format and data type."""
        pass

    @abstractmethod
    def _read(self):
        """Read pixel data from input image."""
        pass


class BrightfieldMemoryBackend(BrightfieldBackend):
    """Brightfield image backend using in-memory `Pillow` reader.

    Parameters
    ----------
    path : str or Path
        Path to the input image file.
    mpp : None | float
        User-provided microns per pixel.
    """

    config: BrightfieldBackendConfig = BRIGHTFIELD_MEMORY_CONFIG

    def _resolution(self) -> None:
        self.mpp_x = self.mpp
        self.mpp_y = self.mpp

    def _validate(self) -> None:
        """Validate input image format.

        Raises
        ------
        ValueError
            If the file does not have a supported format extension.
        """
        if self.l is not None:
            raise ValueError("Level cannot be set when using BrightfieldMemoryBackend.")

        if not self.config.supports_format(self.path):
            raise ValueError(
                "Invalid image format. " + f"Must be one of {', '.join(self.formats)}"
            )

    def _read(self) -> None:
        """Read pixels from validated input image data with `PIL`."""
        self.pyramidal = False

        if self.path.suffix == ".npy":
            data = np.load(self.path)
        else:
            data = np.array(Image.open(self.path).convert("RGB"))

        if data.ndim != 3:
            raise ValueError(
                "BrightfieldMemoryBackend only supports 3-dimensional inputs."
            )

        self.l = None
        self.h, self.w, self.c = data.shape

        if self.c != 3:
            raise ValueError(
                "BrightfieldMemoryBackend is meant for 3-channel RGB images."
            )

        self.data = data

    def __getitem__(self, key: IndexKey) -> Numeric | NDArray[Numeric]:
        """Direct indexing of the underlying pixel data.

        Parameters
        ----------
        key : IndexKey
            A slice, ellipsis, or any valid index into numpy ndarray.

        Returns
        -------
        Numeric | NDArray[Numeric]
            Subregion of brightfield image.
        """
        return self.data[key]


class BrightfieldTiffBackend(BrightfieldBackend):
    """Brightfield image backend using `tifffile` for zarr-backed TIFF formats.

    Parameters
    ----------
    path : str or Path
        Path to the input image file.
    mpp : None | float
        User-provided microns per pixel.
    level : None | int | str
        User-provided level to read.
    """

    config: BrightfieldBackendConfig = BRIGHTFIELD_TIFF_CONFIG

    def _validate(self) -> None:
        """Validate input image format.

        Raises
        ------
        ValueError
            If the file does not have a supported format extension.
        ValueError
            If the file specifies a trestle TIFF.
        ValueError
            If the file species a non-RGB or non-YCBCR photometric image.
        """
        self.backend = "tifffile"
        if not self.config.supports_format(self.path):
            raise ValueError(
                "Invalid image format. " + f"Must be one of {', '.join(self.formats)}"
            )

        with tifffile.TiffFile(self.path) as tiff:
            if copyright_tag := tiff.pages[0].tags.get(33432):
                if "trestle" in copyright_tag._str().lower():
                    # Note: We do not support trestle `tif` files in the
                    # BrightfieldTiffBackend as tiles overlap/aren't stitched
                    # together properly.
                    raise ValueError(
                        "'BrightfieldTiffBackend' does not support trestle "
                        + "`.tif` images."
                    )

            if tiff.pages[0].photometric not in [
                tifffile.PHOTOMETRIC.RGB,
                tifffile.PHOTOMETRIC.YCBCR,
            ]:
                raise ValueError(
                    "'BrightfieldTiffBackend' expects RGB and YCBCR images."
                )

    def _resolution(self) -> None:
        """Extract resolution info from `tifffile` supported formats."""
        with tifffile.TiffFile(self.path) as tiff:
            self.dtype = tiff.pages[0].dtype

            # TO-DO: It may be beneficial to add warnings or errors if an
            # extension is present but the extension is not validated in the
            # parsed header metadata.
            match self.path.suffix.lower():
                case ".ndpi":
                    if tiff.is_ndpi:
                        self._resolution_tags(tiff)
                case ".tiff" | ".tif":
                    if tiff.is_ome:
                        self._resolution_ome_tiff(tiff)
                    if tiff.is_philips:
                        self._resolution_philips(tiff)
                case ".scn":
                    if tiff.is_scn:
                        self._resolution_tags(tiff)
                case ".svs":
                    self._resolution_svs(tiff)

    def _resolution_ome_tiff(self, tiff: TiffFile) -> None:
        """Extract resolution info from `ome.tiff` images."""
        xml_str = tiff.pages[0].description
        root = ET.fromstring(xml_str)
        pixels = root.find(".//{*}Pixels")
        if pixels is not None:
            x = float(pixels.get("PhysicalSizeX", 0))
            y = float(pixels.get("PhysicalSizeY", 0))
            x_unit = pixels.get("PhysicalSizeXUnit", "")
            y_unit = pixels.get("PhysicalSizeYUnit", "")

            if x_unit in ["µm", "cm"] and y_unit in ["µm", "cm"]:
                if x_unit == "cm":
                    x = 1e4 / x

                if y_unit == "cm":
                    y = 1e4 / y

                self.mpp_x = x
                self.mpp_y = y

    def _resolution_philips(self, tiff: TiffFile) -> None:
        """Extract resolution info from philips `.tiff` images."""
        xml_str = tiff.pages[0].description
        root = ET.fromstring(xml_str)
        mmpp = root.find(".//Attribute[@Name='DICOM_PIXEL_SPACING']").text
        if mmpp is not None:
            mmpp_y, mmpp_x = map(float, mmpp.strip('"').split('" "'))
            self.mpp_x = 1000 * mmpp_x
            self.mpp_y = 1000 * mmpp_y

    def _resolution_svs(self, tiff: TiffFile) -> None:
        """Extract resolution metadata from `.svs` images."""
        meta_tag = tiff.pages[0].description
        if match := re.search(r"MPP\s*=\s*([0-9.]+)", meta_tag):
            self.mpp_x = self.mpp_y = float(match.group(1))

    def _resolution_tags(self, tiff: TiffFile) -> None:
        """Extract resolution metadata from `.scn` and `.ndpi` images."""
        page = max(tiff.pages, key=lambda p: p.shape)
        x = page.tags.get("XResolution")
        y = page.tags.get("YResolution")

        self.mpp_x = 1e4 / x.value[0] if x else None
        self.mpp_y = 1e4 / y.value[0] if y else None

    def _read(self) -> None:
        """Read pixels from validated input image data with `tifffile`."""
        if self.path.suffix == ".scn":
            # Note: `.scn` images store a macro image in the first series so we
            # manually set the series here. We may want to build all the format
            # specific readers as methods for clarity.
            store = tifffile.imread(self.path, mode="r", aszarr=True, series=1)
        else:
            store = tifffile.imread(self.path, mode="r", aszarr=True)

        self.pyramidal = store.is_multiscales
        self.data = zarr.open(store, mode="r")

        if self.pyramidal:
            self.levels_keys = []
            self.levels_shape = {}
            for key in sorted(self.data.keys()):
                self.levels_keys.append(key)
                self.levels_shape[key] = self.data[key].shape

            if self.l is None:
                self.l = "0"

            if self.l not in self.levels_keys:
                raise ValueError(
                    f"The provided level ({self.l}) was not found. "
                    + f"Available levels include: {self.levels_keys}"
                )

            self.h, self.w, self.c = self.data[self.l].shape
        else:
            self.h, self.w, self.c = self.data.shape

    def __getitem__(self, key: IndexKey) -> Numeric | NDArray[Numeric]:
        """Direct indexing of the underlying pixel data.

        Parameters
        ----------
        key : IndexKey
            A slice, ellipsis, or any valid index into numpy ndarray.

        Returns
        -------
        Numeric | NDArray[Numeric]
            Subregion of brightfield image.
        """
        if self.pyramidal:
            result = self.data[self.l][key]
        else:
            result = self.data[key]

        if result.size == 1:
            # Note: Single values indexed from `zarr` arrays are returned
            # as arrays. Therefore, we manually check the final size and
            # return scalar values to maintain numpy-like indexing.
            return result.item()

        return np.array(result)


class BrightfieldOpenslideBackend(BrightfieldBackend):
    """Brightfield image backend using `openslide-python` for certain formats.

    Parameters
    ----------
    path : str or Path
        Path to the input image file.
    mpp : None | float
        User-provided microns per pixel.
    """

    config: BrightfieldBackendConfig = BRIGHTFIELD_OPENSLIDE_CONFIG

    def _validate(self) -> None:
        """Validate input image format and `openslide` installation.

        Raises
        ------
        ModuleNotFoundError
            If `openslide` installation is not found.
        ValueError
            If the file does not have a supported format extension.
        """
        self.backend = "openslide"
        if importlib.util.find_spec(self.backend) is None:
            raise ModuleNotFoundError(
                f"'{self.backend}' was not found. Please install to work "
                + f"with {', '.join(self.formats)} formats."
            )

        if not self.config.supports_format(self.path):
            raise ValueError(
                "Invalid image format. " + f"Must be one of {', '.join(self.formats)}"
            )

    def _resolution(self) -> None:
        """Extract resolution info from `openslide` supported formats."""
        import openslide

        with openslide.OpenSlide(self.path) as openslide:
            if "openslide.mpp-x" in openslide.properties:
                self.mpp_x = float(openslide.properties["openslide.mpp-x"])

            if "openslide.mpp-y" in openslide.properties:
                self.mpp_y = float(openslide.properties["openslide.mpp-y"])

            # Note: OpenSlide will fill mpp metadata with large values in some
            # cases, like generic tiffs. For now, we just check for nonsensical
            # values and convert mpp back to the empty placeholder. Most slides
            # are read at 20x/40x magnification so things above 10 are highly
            # unlikely since most if not all slides are imaged at mpp < 1.
            if self.mpp_x > 10.0:
                self.mpp_x = None

            if self.mpp_y > 10.0:
                self.mpp_y = None

    def _read(self) -> None:
        """Read pixels from validated input image data with `openslide`."""
        import openslide

        self.data = openslide.OpenSlide(self.path)

        # Note: Openslide only supports 8-bit RGB/A slides.
        self.dtype = np.uint8
        self.pyramidal = True
        self.c = 3

        self.levels = self.data.level_count
        self.levels_keys = list(range(self.levels))

        # Note: Openslide returns (width, height) so we swap the dimensions
        # here to maintain consistency across different brightfield backends.
        self.levels_shape = {
            k: (v[1], v[0], 3)
            for k, v in zip(self.levels_keys, self.data.level_dimensions)
        }

        if self.l is None:
            self.l = 0

        if self.l not in self.levels_keys:
            raise ValueError(
                f"The provided level ({self.l}) was not found. "
                + f"Available levels include: {self.levels_keys}"
            )

        self.h, self.w, _ = self.levels_shape[self.l]

    def __getitem__(self, key: IndexKey) -> Numeric | NDArray[Numeric]:
        """Direct indexing of the underlying pixel data.

        Parameters
        ----------
        key : IndexKey
            A slice, ellipsis, or any valid index into numpy ndarray.

        Returns
        -------
        Numeric | NDArray[Numeric]
            Subregion of brightfield image.
        """

        def _expand_ellipsis(key, ndim=3):
            if not isinstance(key, tuple):
                key = (key,)
            if Ellipsis not in key:
                return key + (slice(None),) * (ndim - len(key))
            n_ellipsis = key.count(Ellipsis)
            if n_ellipsis > 1:
                raise IndexError("An index can only have a single ellipsis ('...')")
            idx = key.index(Ellipsis)
            num_missing = ndim - (len(key) - 1)
            if num_missing < 0:
                raise IndexError("Too many indices for image dimensions")
            return key[:idx] + (slice(None),) * num_missing + key[idx + 1 :]

        def _normalize_index(k, max_val):
            if isinstance(k, int):
                if k < 0:
                    k += max_val
                if not (0 <= k < max_val):
                    raise IndexError(
                        f"Index {k} out of bounds for dimension size {max_val}"
                    )
                return slice(k, k + 1), 1
            elif isinstance(k, slice):
                start, stop, step = k.indices(max_val)
                if step == 0:
                    raise ValueError("Slice step cannot be zero")
                length = max(0, (stop - start + (step - 1)) // step)
                return slice(start, stop, step), length
            else:
                raise TypeError("Index must be int or slice")

        key = _expand_ellipsis(key, ndim=3)

        if len(key) > 3:
            raise IndexError(
                "Indexing must be in the form [y], [y, x], or [y, x, channel]"
            )

        while len(key) < 3:
            key = key + (None,)

        y_key, x_key, channel = key

        y_slice, y_len = _normalize_index(y_key, self.h)
        x_slice, x_len = _normalize_index(x_key, self.w)

        y_start, y_stop, y_step = y_slice.start, y_slice.stop, y_slice.step
        x_start, x_stop, x_step = x_slice.start, x_slice.stop, x_slice.step

        region_width = x_stop - x_start
        region_height = y_stop - y_start

        if region_width <= 0 or region_height <= 0:
            raise IndexError("Empty or invalid slice requested")

        region = self.data.read_region(
            (x_start, y_start), self.l, (region_width, region_height)
        )
        region = region.convert("RGB")
        region = np.array(region)

        region = region[::y_step, ::x_step]
        if channel is not None:
            region = region[..., channel]

        return region.item() if region.size == 1 else np.squeeze(region)



from dataclasses import dataclass

DATA: str = "src/tests/data/"


@dataclass(frozen=True)
class RgbBmp:
    path: str = DATA + "bmp/rgb.bmp"
    mpp_x: None | float = None
    mpp_y: None | float = None
    width: int = 621
    height: int = 621
    channels: int = 3


@dataclass(frozen=True)
class RgbJpeg:
    path: str = DATA + "jpeg/rgb.jpeg"
    mpp_x: None | float = None
    mpp_y: None | float = None
    width: int = 621
    height: int = 621
    channels: int = 3


@dataclass(frozen=True)
class RgbNpy:
    path: str = DATA + "npy/rgb.npy"
    mpp_x: None | float = None
    mpp_y: None | float = None
    width: int = 621
    height: int = 621
    channels: int = 3


@dataclass(frozen=True)
class RgbPbm:
    path: str = DATA + "pbm/rgb.pbm"
    mpp_x: None | float = None
    mpp_y: None | float = None
    width: int = 621
    height: int = 621
    channels: int = 3


@dataclass(frozen=True)
class RgbPng:
    path: str = DATA + "png/rgb.png"
    mpp_x: None | float = None
    mpp_y: None | float = None
    width: int = 621
    height: int = 621
    channels: int = 3


@dataclass(frozen=True)
class RgbTga:
    path: str = DATA + "tga/rgb.tga"
    mpp_x: None | float = None
    mpp_y: None | float = None
    width: int = 621
    height: int = 621
    channels: int = 3


@dataclass(frozen=True)
class RgbWebp:
    path: str = DATA + "webp/rgb.webp"
    mpp_x: None | float = None
    mpp_y: None | float = None
    width: int = 621
    height: int = 621
    channels: int = 3



import pytest
from bamboo.image import Brightfield
from bamboo.tile import Tiles
from bamboo.utils import Map, ProcessMap, ThreadMap, apply_map
from tests._fixtures.slides import SlideSvs


def test_map_basic():
    image = Brightfield(SlideSvs.path, level="2")
    results = Map(lambda tile: tile.shape, Tiles(image, 128, 128))
    for shape in results:
        assert shape == (128, 128, 3)


def test_parallel_map_threads():
    image = Brightfield(SlideSvs.path, level="2")
    results_a = Map(lambda tile: tile.shape, Tiles(image, 128, 128))
    results_b = ThreadMap(lambda tile: tile.shape, Tiles(image, 128, 128), n_jobs=2)
    for a, b in zip(results_a, results_b):
        assert a == b


def test_parallel_map_processes():
    image = Brightfield(SlideSvs.path, level="2")
    results_a = Map(lambda tile: tile.shape, Tiles(image, 128, 128))
    results_b = ProcessMap(lambda tile: tile.shape, Tiles(image, 128, 128), n_jobs=2)
    for a, b in zip(results_a, results_b):
        assert a == b


def test_apply_map_functionality():
    def _func(i):
        return i * 2

    single_map = apply_map(_func, range(10), n_jobs=1, silent=True)
    process_map = apply_map(_func, range(10), n_jobs=2, silent=True)
    thread_map = apply_map(_func, range(10), n_jobs=2, prefer="threads", silent=True)

    for result in [single_map, process_map, thread_map]:
        for i, val in enumerate(result):
            assert val == i * 2



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

import os
import shutil
import numpy as np
import tempfile

from anci import Arg, cmd, base
from pathlib import Path

from bamboo.tools._resize import resize_mask_memmap, resize_mask_zarr
from bamboo.writer import NumpyWriter, SharedMemoryWriter, MemmapWriter, ZarrWriter

SUPPORTED_MODELS: set[str] = {"grandqc"}


@base("brightfield", "nn")
def brightfield_nn():
    """Neural networks for brightfield imaging."""
    pass


@cmd("brightfield", "nn", "semantic")
def brightfield_nn_semantic(
    path: Arg[Path, "Path to brightfield image."],
    model: Arg[str, "Model name."],
    output: Arg[Path, "Output path."],
    mmap: Arg[bool, "If True, use a memory map when writing .npy outputs."] = True,
    n_jobs: Arg[int, "Number of parallel jobs."] = 1,
    prefer: Arg[str, "Parallel backend hint."] = None,
    resample: Arg[str, "Resampling filter to use when resizing tiles."] = "bilinear",
    resize: Arg[str, "Resize segmentation mask to original image size."] = False,
    device: Arg[str, "Device to run model on."] = "cpu",
    silent: Arg[bool, "Suppress printed messages."] = False,
):
    """Generate a semantic segmentation mask from an input brightfield image."""

    from bamboo.image import Brightfield
    from bamboo.models import Semantic
    from bamboo.tile import Tiles

    image = Brightfield(path)

    if prefer not in ["processes", "threads", None]:
        raise ValueError(
            f"''--prefer' {prefer} is not valid. "
            "Must be one of 'processes', 'threads', or None."
        )

    if not any(
        [device.startswith("cpu"), device.startswith("cuda"), device.startswith("mps")]
    ):
        raise ValueError(
            f"The '--device' {device} is not valid. "
            "Currently supported devices include: 'cpu', 'cuda', 'mps'"
        )

    if model not in SUPPORTED_MODELS:
        raise ValueError(
            f"The '--model' {model} is not valid. "
            f"Currently supported models include: {SUPPORTED_MODELS}"
        )

    match model:
        case "grandqc":
            from bamboo.models.zoo import GrandQC

            model = GrandQC(device=device)

    match output.suffix:
        case ".npy":
            if mmap:
                writer = MemmapWriter(dtype=model.dtype, path=output)
            else:
                if n_jobs == 1:
                    writer = NumpyWriter(dtype=model.dtype)
                else:
                    writer = SharedMemoryWriter(dtype=model.dtype)
        case ".zarr":
            writer = ZarrWriter(dtype=model.dtype, path=output)
        case _:
            raise ValueError(f"The --output {output} must end with .npy or .zarr.")

    tiles = Tiles(
        image,
        model.tile_width,
        model.tile_height,
        model.mpp,
        resample=resample,
        overlap_x=0,
        overlap_y=0,
    )

    engine = Semantic(
        model,
        tiles,
        writer=writer,
        n_jobs=n_jobs,
        prefer=prefer,
    )

    results = engine.run(silent=silent)

    if output.suffix == ".npy" and not mmap:
        np.save(output, results)
        writer.cleanup()



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

import numpy as np
import zarr


def resize_nearest(img: np.ndarray, new_shape: tuple[int, int]) -> np.ndarray:
    """Resize array using nearest-neighbor interpolation.

    Parameters
    ----------
    img : np.ndarray
        Input image of shape (H, W) or (H, W, C)
    new_shape : (new_H, new_W)
        Desired output shape.

    Returns
    -------
    np.ndarray
        Resized image of shape (new_H, new_W[, C])
    """
    h, w = img.shape[:2]
    new_h, new_w = new_shape

    # Compute scale factors
    row_scale = h / new_h
    col_scale = w / new_w

    # Compute source indices
    row_idx = (np.arange(new_h) * row_scale).astype(int)
    col_idx = (np.arange(new_w) * col_scale).astype(int)

    # Clamp just in case (avoid out-of-bounds on last pixel)
    row_idx = np.clip(row_idx, 0, h - 1)
    col_idx = np.clip(col_idx, 0, w - 1)

    # Use advanced indexing for fast vectorized sampling
    if img.ndim == 3:
        return img[row_idx[:, None], col_idx[None, :], :]
    else:
        return img[row_idx[:, None], col_idx[None, :]]


def resize_mask_memmap(
    input_path: str,
    output_path: str,
    target_shape: tuple[int, ...],
    row_block: int = 1000,
    col_block: int = 1000,
) -> np.memmap:
    """Resize memory-mapped segmentation mask using block-wise nearest-neighbor indexing.

    Parameters
    ----------
    input_path : str
        Path to original .npy mask.
    output_path : str
        Path to save resized mask.
    target_shape : tuple[int, int]
        Resize to specified height and width.
    row_block : int
        Number of output rows per block.
    col_block : int
        Number of output columns per block.
    """
    # Load original mask as memmap
    orig_mask = np.load(input_path, mmap_mode="r")
    old_H, old_W = orig_mask.shape
    new_H, new_W = target_shape

    # Create output memmap
    resized_mask = np.lib.format.open_memmap(
        output_path, dtype=orig_mask.dtype, mode="w+", shape=(new_H, new_W)
    )

    # Precompute nearest-neighbor indices
    row_idx = (np.linspace(0, old_H - 1, new_H)).astype(int)
    col_idx = (np.linspace(0, old_W - 1, new_W)).astype(int)

    # Process in row + column blocks
    for row_start in range(0, new_H, row_block):
        row_end = min(row_start + row_block, new_H)
        rows = row_idx[row_start:row_end]

        for col_start in range(0, new_W, col_block):
            col_end = min(col_start + col_block, new_W)
            cols = col_idx[col_start:col_end]

            # Slice original mask and write to output memmap
            resized_mask[row_start:row_end, col_start:col_end] = orig_mask[
                rows[:, None], cols[None, :]
            ]

    # Flush changes to disk
    resized_mask.flush()
    return resized_mask


def resize_mask_zarr(
    input_path: str,
    output_path: str,
    target_shape: tuple[int, int],
    row_block: int = 1000,
    col_block: int = 1000,
) -> zarr.Array:
    """
    Resize a Zarr segmentation mask array using block-wise nearest-neighbor interpolation.

    Parameters
    ----------
    input_path : str
        Path to source .zarr directory.
    output_path : str
        Path to destination .zarr directory.
    target_shape : tuple[int, int]
        Desired (height, width) of the output mask.
    row_block : int
        Number of output rows per block.
    col_block : int
        Number of output columns per block.

    Returns
    -------
    zarr.Array
        Resized Zarr array on disk.
    """
    # Open input Zarr in read-only mode
    src = zarr.open(input_path, mode="r")
    old_H, old_W = src.shape
    dtype = src.dtype

    new_H, new_W = target_shape

    # Create destination Zarr array
    dst = zarr.open(
        output_path,
        mode="w",
        shape=(new_H, new_W),
        dtype=dtype,
        chunks=(row_block, col_block),
    )

    # Precompute nearest-neighbor index maps
    row_idx = np.linspace(0, old_H - 1, new_H).astype(int)
    col_idx = np.linspace(0, old_W - 1, new_W).astype(int)

    # Iterate through output blocks
    for row_start in range(0, new_H, row_block):
        row_end = min(row_start + row_block, new_H)
        rows = row_idx[row_start:row_end]

        for col_start in range(0, new_W, col_block):
            col_end = min(col_start + col_block, new_W)
            cols = col_idx[col_start:col_end]

            # Read block from source and write to destination
            dst[row_start:row_end, col_start:col_end] = src[
                rows[:, None], cols[None, :]
            ]

    return dst



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

from ._base import Tile
from ._generic import GenericTile
from ._iter import Tiles
from ._rgb import RGBTile

__all__ = ["Tile", "GenericTile", "Tiles", "RGBTile"]



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

import numpy as np

from typing import Callable

from bamboo.tile import Tiles, Tile
from bamboo.tools._resize import resize_nearest
from bamboo.types import Array
from bamboo.utils import apply_map
from bamboo.writer import ArrayWriter, NumpyWriter, ZarrWriter, inject_writer


class Semantic:
    """Perform semantic segmentation across tiled images.

    Parameters
    ----------
    func : Callable[Tile, np.ndarray]
        A semantic segmentation model that takes as input a tile and
        returns as output a numpy ndarray with shape (h, w).
    tiles : Tiles
        An iterable of tiles that are scaled to a size compatible with
        the provided semantic segmentation model.
    writer : ArrayWriter
        Array writer backend. If `str`, then one of 'numpy', 'shared', or
        'memmap'. Otherwise, a valid ArrayWriter derived class.
    n_jobs : None | int
        Number of parallel jobs.
    prefer : None | str
        Either None, 'processes, or 'threads'. If n_jobs and prefer are both
        None, then no parallelism will be used (n_jobs == 1). If n_jobs > 1 and
        prefer is None, then 'processes' will be used. If prefer is 'processes'
        and n_jobs is None, all available processes will be used.

    Example
    -------
    This example demonstrates the general usage of the `Semantic` runner class.
    A `Brightfield` image is loaded and then a `Semantic` segmentation runner is
    initialized using a `GrandQC` model, a `Tiles` tile generator at a specific
    tile size and microns per pixel resolution, and with the results are written
    to a memory mapped array using `MemmapWriter`.

    >>> from bamboo.image import Brightfield
    >>> from bamboo.tile import Tiles
    >>> from bamboo.writer import MemmapWriter
    >>> from bamboo.models import Semantic
    >>> from bamboo.models.zoo import GrandQC

    >>> image = Brightfield("slide.svs")

    >>> runner = Semantic(
    ...     GrandQC(device="cpu"),
    ...     Tiles(image, tile_width=512, tile_height=512, mpp=1.5),
    ...     writer=MemmapWriter(dtype=np.uint8, path="array.npy"),
    ...     n_jobs=8,
    ...     prefer="processes",
    ... )

    >>> output = runner.run()
    """

    def __init__(
        self,
        func: Callable[[Tile], np.ndarray],
        tiles: Tiles,
        writer: ArrayWriter,
        n_jobs: int | None = None,
        prefer: str | None = None,
    ):
        if prefer == "processes" and isinstance(writer, NumpyWriter):
            raise ValueError(
                "If prefer is 'processes', then writer must one of: "
                "`SharedMemoryWriter` or `MemmapWriter`."
            )

        self.func = func
        self.tiles = tiles
        self.n_jobs = n_jobs
        self.prefer = prefer
        self.writer = writer
        self.dtype = writer.dtype

        if isinstance(writer, ZarrWriter):
            writer.chunks = (tiles.crop_h, tiles.crop_w)

        if tiles.scale is None and writer.shape is None:
            raise ValueError(
                "If 'tiles' has no scale or microns per pixel information,"
                "then `writer` must have shape set prior to running `Semantic`."
            )

        if writer.shape is None:
            writer.shape = (tiles.image_h, tiles.image_w)

        if self.tiles.mode != "coords":
            self.tiles.mode = "coords"

    def run(self, silent: bool = False) -> Array:
        self.writer.create()

        @inject_writer(self.writer)
        def _worker(output: Array, data: tuple[tuple[int, ...], Tile]) -> np.ndarray:
            (x, y, w, h, overlap_w, overlap_h), tile = data

            processed = self.func(tile).astype(output.dtype)
            if processed.shape[:2] != (h, w):
                processed = resize_nearest(processed, (h, w))

            output[y + overlap_h : y + h, x + overlap_w : x + w] = processed[
                overlap_h:, overlap_w:
            ]

        apply_map(
            _worker,
            self.tiles,
            self.n_jobs,
            self.prefer,
            silent,
        )

        return self.writer.output


class SemanticStats:
    """Run a semantic segmentation model and compute statistics on each tile mask.

    Parameters
    ----------
    func : Callable[[Tile], np.ndarray]
        A semantic segmentation function that takes a tile and returns a mask (H, W).
    counter : Callable[np.ndarray, dict]
        A function that computes statistics using the generated mask.
    tiles : Tiles
        Iterable of tiles compatible with `func`.
    n_jobs : None | int, optional
        Number of parallel jobs for processing tiles.
    prefer : None | str, optional
        Parallel backend preference ('processes' or 'threads').

    Returns
    -------
    list[dict]
        Each entry corresponds to a tile and includes its coordinates and statistics.
    """

    def __init__(
        self,
        func: Callable[[Tile], np.ndarray],
        counter: Callable[np.ndarray, dict],
        tiles: Tiles,
        n_jobs: int | None = None,
        prefer: str | None = None,
    ):
        self.func = func
        self.counter = counter
        self.tiles = tiles
        self.n_jobs = n_jobs
        self.prefer = prefer

        if self.tiles.mode != "coords":
            self.tiles.mode = "coords"

    def run(self, silent: bool = False):
        """Run the segmentation and compute statistics for each tile."""
        results = []

        def _worker(data: tuple[tuple[int, ...], Tile]):
            (x, y, w, h, overlap_w, overlap_h), tile = data
            mask = self.func(tile)
            stats = self.counter(mask)
            return {"stats": stats, "coords": [x, y, w, h, overlap_w, overlap_h]}

        results = apply_map(
            _worker,
            self.tiles,
            self.n_jobs,
            self.prefer,
            silent,
        )

        return results



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

import urllib.request
import urllib.parse
import http.cookiejar
import re

from pathlib import Path
from tqdm import tqdm
from urllib.error import URLError, HTTPError


MODELS: dict = {
    "grandqc": {"source": "gdrive", "identifier": "1B_uRQW-LWUUaIxowibaOCFFibFDwp_u6"}
}


def download_gdrive(identifier: str, destination: str | Path) -> None:
    """Download a file from google drive using its file identifier.

    Parameters
    ----------
    identifier : str
        Google drive file identifier specifying a publicly available file.
    destination : str or Path
        Path to save the downloaded file.

    Raises
    ------
    RuntimeError
        If the download fails or cannot be retrieved from Google Drive.
    URLError
        If the network request fails due to connectivity or URL issues.
    HTTPError
        If the HTTP request returns a non-successful status code.
    """
    URL = "https://docs.google.com/uc?export=download"
    destination = Path(destination)

    session = urllib.request.build_opener(
        urllib.request.HTTPCookieProcessor(http.cookiejar.CookieJar())
    )

    try:
        response = session.open(f"{URL}&id={identifier}")
        html = response.read()
    except (HTTPError, URLError) as e:
        raise RuntimeError(f"Failed to initiate download for {identifier}: {e}")

    try:
        html_text = html.decode("utf-8")
        token_match = re.search(r"confirm=([0-9A-Za-z_]+)", html_text)
        token = token_match.group(1) if token_match else None

        if token:
            response = session.open(f"{URL}&confirm={token}&id={identifier}")
    except UnicodeDecodeError:
        # Already binary content, no token step required
        response = session.open(f"{URL}&id={identifier}")

    total_size = int(response.headers.get("Content-Length", 0))

    with (
        open(destination, "wb") as f,
        tqdm(
            total=total_size,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            desc=f"Downloading {destination.name}",
            ncols=80,
        ) as pbar,
    ):
        while True:
            chunk = response.read(32768)
            if not chunk:
                break
            f.write(chunk)
            pbar.update(len(chunk))


def download_model(name: str, dest: str | Path) -> Path:
    """Download a model from a predefined registry of available models.

    Parameters
    ----------
    name : str
        The model name as defined `MODELS` dictionary.
    dest : str or Path
        Path where the model file should be saved.

    Returns
    -------
    pathlib.Path
        The resolved path to the downloaded model file.

    Raises
    ------
    ValueError
        If the model name is not defined in ``MODELS``.
    NotImplementedError
        If the model source type is unsupported.
    RuntimeError
        If the download process encounters an unrecoverable error.
    """
    if name not in MODELS:
        raise ValueError(f"Unknown model '{name}'")

    model = MODELS[name]
    dest = Path(dest)
    dest.parent.mkdir(parents=True, exist_ok=True)

    if model["source"] == "gdrive":
        download_gdrive(model["identifier"], dest)
    else:
        raise NotImplementedError(f"Unsupported source: {model['source']}")

    return dest



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

import numpy as np

from anci import Arg, cmd, base
from pathlib import Path

from bamboo.types import Array
from bamboo.utils import apply_map

from bamboo.writer import (
    NumpyWriter,
    SharedMemoryWriter,
    MemmapWriter,
    ZarrWriter,
    inject_writer,
)

SUPPORTED_MODELS: set[str] = {"grandqc"}


@base("brightfield", "tile")
def brightfield_nn():
    """Methods for processing, filtering, and writing brightfield tiles,."""
    pass


@cmd("brightfield", "tile", "extract")
def brightfield_tile_extract(
    path: Arg[Path, "Path to brightfield image."],
    output: Arg[Path, "Output path."],
    tile_width: Arg[int, "Tile width."],
    tile_height: Arg[int, "Tile height."],
    mpp: Arg[int, "Collect tiles at the specified microns per pixel."] = None,
    overlap_x: Arg[int, "Horizontal overlap when saving tiles."] = 0,
    overlap_y: Arg[int, "Vertical overlap when saving tiles."] = 0,
    mmap: Arg[bool, "If True, use a memory map when writing .npy outputs."] = True,
    n_jobs: Arg[int, "Number of parallel jobs."] = 1,
    prefer: Arg[str, "Parallel backend hint."] = None,
    resample: Arg[str, "Resampling filter to use when resizing tiles."] = "bilinear",
    silent: Arg[bool, "Suppress printed messages."] = False,
):
    """Extract and write all tiles from a brightfield gigapixel image."""

    from bamboo.image import Brightfield
    from bamboo.tile import Tiles, Tile

    image = Brightfield(path)

    if prefer not in ["processes", "threads", None]:
        raise ValueError(
            f"''--prefer' {prefer} is not valid. "
            "Must be one of 'processes', 'threads', or None."
        )

    match output.suffix:
        case ".npy":
            if mmap:
                writer = MemmapWriter(dtype=np.uint8, path=output)
            else:
                if n_jobs == 1:
                    writer = NumpyWriter(dtype=np.uint8)
                else:
                    writer = SharedMemoryWriter(dtype=np.uint8)
        case ".zarr":
            writer = ZarrWriter(dtype=np.uint8, path=output)
        case _:
            raise ValueError(f"The --output {output} must end with .npy or .zarr.")

    mpp = image.mpp if mpp is None else mpp

    tiles = Tiles(
        image,
        tile_width,
        tile_height,
        mpp,
        resample=resample,
        overlap_x=overlap_x,
        overlap_y=overlap_y,
        mode="enumerate",
    )

    h, w = tiles.hw
    if isinstance(writer, ZarrWriter):
        writer.chunks = (h, w, 3)

    writer.shape = (tiles.n_tiles, h, w, 3)

    @inject_writer(writer)
    def _worker(output: Array, data: tuple[tuple[int, ...], Tile]) -> np.ndarray:
        i, tile = data
        output[i, :h, :w, :3] = tile.buffer[:, :, :3]

    writer.create()

    try:
        apply_map(
            _worker,
            tiles,
            n_jobs,
            prefer,
            silent,
        )

        if output.suffix == ".npy" and not mmap:
            np.save(output, writer.output)
            writer.cleanup()
    except Exception as e:
        writer.cleanup()
        raise RuntimeError("Failed to write tiles to disk") from e


@cmd("brightfield", "tile", "filter")
def brightfield_tile_filter(
    path: Arg[Path, "Path to brightfield image."],
    model: Arg[str, "Filtering model (one of: 'grandqc')."],
    output: Arg[Path, "Output path."],
    tile_width: Arg[int, "Tile width."],
    tile_height: Arg[int, "Tile height."],
    min_tissue_fraction: Arg[float, "Minimum tissue fraction to keep a tile."] = 0.5,
    max_artifact_fraction: Arg[float, "Max artifact fraction to keep a tile."] = 0.0,
    mpp: Arg[int, "Collect tiles at the specified microns per pixel."] = None,
    overlap_x: Arg[int, "Horizontal overlap when saving tiles."] = 0,
    overlap_y: Arg[int, "Vertical overlap when saving tiles."] = 0,
    mmap: Arg[bool, "If True, use a memory map when writing .npy outputs."] = True,
    n_jobs: Arg[int, "Number of parallel jobs."] = 1,
    prefer: Arg[str, "Parallel backend hint."] = None,
    resample: Arg[str, "Resampling filter to use when resizing tiles."] = "bilinear",
    device: Arg[str, "Device to run model on."] = "cpu",
    silent: Arg[bool, "Suppress printed messages."] = False,
):
    """Filter and write all tiles from a brightfield gigapixel image."""

    from bamboo.image import Brightfield
    from bamboo.tile import Tiles, Tile
    from bamboo.models import SemanticStats

    image = Brightfield(path)

    if prefer not in ["processes", "threads", None]:
        raise ValueError(
            f"''--prefer' {prefer} is not valid. "
            "Must be one of 'processes', 'threads', or None."
        )

    if not any(
        [device.startswith("cpu"), device.startswith("cuda"), device.startswith("mps")]
    ):
        raise ValueError(
            f"The '--device' {device} is not valid. "
            "Currently supported devices include: 'cpu', 'cuda', 'mps'"
        )

    if model not in SUPPORTED_MODELS:
        raise ValueError(
            f"The '--model' {model} is not valid. "
            f"Currently supported models include: {SUPPORTED_MODELS}"
        )

    match model:
        case "grandqc":
            from bamboo.models.zoo import GrandQC, GrandQCStats

            model = GrandQC(device=device)
            stats = GrandQCStats

    match output.suffix:
        case ".npy":
            if mmap:
                writer = MemmapWriter(dtype=np.uint8, path=output)
            else:
                if n_jobs == 1:
                    writer = NumpyWriter(dtype=np.uint8)
                else:
                    writer = SharedMemoryWriter(dtype=np.uint8)
        case ".zarr":
            writer = ZarrWriter(dtype=np.uint8, path=output)
        case _:
            raise ValueError(f"The --output {output} must end with .npy or .zarr.")

    mpp = image.mpp if mpp is None else mpp

    # Note: To filter out "bad" tiles we first have to apply a model to each tile
    # compute the relevant statistics and then re-iterate through the retained set
    # of coordinates once more for writing. As such the writer shouldn't be created
    # until the filtering step has been performed.

    tiles = Tiles(
        image,
        model.tile_width,
        model.tile_height,
        model.mpp,
        resample=resample,
        overlap_x=0,
        overlap_y=0,
        mode="coords",
    )

    try:
        engine = SemanticStats(model, stats, tiles, n_jobs, prefer)
        results = engine.run(silent)
    except Exception as e:
        raise RuntimeError(f"Filtering failed using {model} model.") from e

    exclude_coords = []
    for r in results:
        if any(
            [
                r["stats"]["TISSUE"] < min_tissue_fraction,
                r["stats"]["ARTIFACT"] > max_artifact_fraction,
            ]
        ):
            exclude_coords.append(r["coords"])

    # Note: We construct new tiles at the requested size and resolution, then
    # drop the tiles that not pass the quality control metrics. This seems a
    # bit redundant generating tiles twice. An alternative would be to add a
    # method to `Tiles` to regenerate coordinates at a new size given a set of
    # coordinates to exclude.
    tiles = Tiles(
        image,
        tile_width,
        tile_height,
        mpp,
        resample=resample,
        overlap_x=overlap_x,
        overlap_y=overlap_y,
        mode="coords",
    )

    n_tiles = tiles.n_tiles
    tiles.exclude(exclude_coords)

    if not silent:
        print(f"Excluded a total of {n_tiles - tiles.n_tiles} after filtering.")

    # Note: We switch back to enumeration here since we don't need coordinate
    # information to write to tiles. With that said, users may want coordinate
    # information to map tiles back to original images after downstream analysis
    # so maybe come back to here when that arises. If we simplified the API to
    # just work with zarrs then we could store coordinate information in the meta
    # data.
    tiles.mode = "enumerate"

    h, w = tiles.hw
    if isinstance(writer, ZarrWriter):
        writer.chunks = (h, w, 3)

    writer.shape = (tiles.n_tiles, h, w, 3)

    @inject_writer(writer)
    def _worker(output: Array, data: tuple[tuple[int, ...], Tile]) -> np.ndarray:
        i, tile = data
        output[i, :h, :w, :3] = tile.buffer[:, :, :3]

    writer.create()

    try:
        apply_map(
            _worker,
            tiles,
            n_jobs,
            prefer,
            silent,
        )

        if output.suffix == ".npy" and not mmap:
            np.save(output, writer.output)
            writer.cleanup()
    except Exception as e:
        writer.cleanup()
        raise RuntimeError("Failed to write tiles to disk") from e



# Copyright (c) 2025, Tom Ouellette
# Licensed under the Apache 2.0 License

import os
import numpy as np
import torch

from enum import Enum
from numpy.typing import NDArray
from pathlib import Path

from bamboo.models import CACHE_DIR
from bamboo.tile import RGBTile
from bamboo.models._manager import download_model


class GrandQCClasses(Enum):
    """Class labels for GrandQC semantic segmentation."""

    TISSUE = 1
    FOLDS = 2
    SPOTS = 3
    MARKINGS = 4
    ARTIFACTS = 5
    OUT_OF_FOCUS = 6
    BACKGROUND = 7


_MAX_CLASS_VALUE = max(c.value for c in GrandQCClasses)


class GrandQC:
    """A semantic segmentation model for detecting histology image artifacts.

    Parameters
    ----------
    device : str | torch.device
        A valid torch.device (e.g. "cpu", torch.device("cuda:0"), etc).
    path : str | Path
        Optional path to weights if outside BAMBOO_CACHE.

    References
    ----------
    .. [1] Zhilong Weng et al. "GrandQC: A comprehensive solution to quality
       control problem in digital pathology". Nature Communications. 2024.

    License
    -------
    Creative Commons Attribution Non Commercial Share Alike 4.0 International
    """

    # Weights defines the path to the torchscript model
    weights: str = f"{CACHE_DIR}/model.grandqc.pt"

    # Normalization/scaling coefficients
    mean: list[float, ...] = [0.485, 0.456, 0.406]
    std: list[float, ...] = [0.229, 0.224, 0.225]

    # Tile specifications
    tile_width: int = 512
    tile_height: int = 512
    mpp: float = 1.5
    dtype: np.dtype = np.uint8

    # Note: Storing our loaded model as a class-level attribute allows
    # us to load the model only once per process instead of each time.
    _model: torch.nn.Module = None

    def __init__(self, device: str | torch.device = "cpu", path: str | Path = None):
        if not os.path.isfile(self.weights):
            print("Weights not found. Downloading `grandqc` weights to cache.")
            download_model("grandqc", self.weights)

        self.device = device

    def _preprocess(self, tile: RGBTile) -> torch.Tensor:
        x = (tile[:] / 255.0) - np.array(self.mean)
        x /= np.array(self.std)
        x = torch.from_numpy(x).permute(2, 0, 1)
        return x.unsqueeze(0).float().to(self.device)

    def _run(self, tile: RGBTile) -> NDArray[np.uint8]:
        if GrandQC._model is None:
            GrandQC._model = torch.jit.load(self.weights).to(self.device)
            GrandQC._model.eval()

        return (
            GrandQC._model(self._preprocess(tile))
            .detach()
            .cpu()
            .numpy()[0]
            .argmax(axis=0)
        )

    __call__ = _run


def GrandQCStats(tile: NDArray[np.uint8]) -> dict[str, float]:
    """Compute GrandQC class stats from input segmentation mask.

    Parameters
    ----------
    tile : np.ndarray
        2D array (H, W) with integer labels corresponding to GrandQCClasses.

    Returns
    -------
    dict[str, float]
        Mapping from class name (string) to frequency.
    """
    if tile.ndim != 2:
        raise ValueError(f"Expected 2D segmentation mask, got shape {tile.shape}")

    counts = np.bincount(tile.ravel(), minlength=_MAX_CLASS_VALUE + 1)
    stats = {cls.name: counts[cls.value] / tile.size for cls in GrandQCClasses}
    stats["ARTIFACT"] = 1 - stats["TISSUE"] - stats["BACKGROUND"]
    return stats



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

import numpy as np

from dataclasses import dataclass

from bamboo.error import ShapeError
from bamboo.tile._base import Tile


@dataclass(frozen=True)
class GenericTile(Tile):
    """A base class for a two-dimensional in-memory tile of image subregion.

    Attributes
    ----------
    buffer : np.ndarray
        Two-dimensional pixel container with optional last index channel.
    dtype : np.dtype
        Data type of buffer elements (subpixels).
    shape : Tuple[int, int, int]
        Height, width, and channels for the tile.
    h : int
        Height of the tile.
    w : int
        Width of the tile.
    c : int
        Number of channels in the tile.
    """

    def __post_init__(self):
        """Buffer validation and attribute initialization."""
        if not isinstance(self.buffer, np.ndarray):
            raise TypeError("[GenericTile] 'buffer' must be of type np.ndarray.")

        dtype = self.buffer.dtype
        if not np.any(
            [np.issubdtype(dtype, np.integer), np.issubdtype(dtype, np.floating)]
        ):
            raise TypeError("[GenericTile] 'buffer' must be integer or float type.")

        object.__setattr__(self, "dtype", dtype)

        if len(self.buffer.shape) == 2:
            object.__setattr__(self, "buffer", self.buffer[:, :, None])

        if self.buffer.ndim != 3:
            raise ShapeError("[GenericTile] 'buffer' must be 2 or 3 dimensions.")

        shape = self.buffer.shape
        object.__setattr__(self, "shape", shape)
        object.__setattr__(self, "h", shape[0])
        object.__setattr__(self, "w", shape[1])
        object.__setattr__(self, "c", shape[2])



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

from pathlib import Path

from numpy.typing import NDArray
from bamboo.types import IndexKey, Numeric
from bamboo.tile import RGBTile

from bamboo.backend._brightfield import (
    BrightfieldTiffBackend,
    BrightfieldOpenslideBackend,
    BrightfieldMemoryBackend,
    BRIGHTFIELD_TIFF_CONFIG,
    BRIGHTFIELD_OPENSLIDE_CONFIG,
    BRIGHTFIELD_MEMORY_CONFIG,
)


class Brightfield:
    """A class for interacting with brightfield images.

    This class provides a simple interface for interacting with single-level
    in-memory or multi-level memory-mapped/on-disk brightfield RGB images.

    Parameters
    ----------
    path : str or Path
        Path to the input image file.
    mpp : None | float
        User-provided microns per pixel.
    level : None | int | str
        User-provided level to read.
    backend : None | str
        Manually set the backend ('memory', 'tifffile', 'openslide'). The
        backend will automatically selected if not provided.

    Attributes
    ----------
    backend : BrightfieldBackend
        A backend providing access to pixel-level data.
    """

    def __init__(
        self,
        path: str | Path,
        mpp: None | float = None,
        level: None | str | int = None,
        backend: None | str = None,
    ):
        if not isinstance(path, (str, Path)):
            raise ValueError("'path' must be a str or pathlib.Path.")

        if not (path := Path(path)).exists():
            raise FileNotFoundError("'path' does not exist.")

        if path.is_dir():
            raise OSError("'path' must be a file not a directory.")

        if backend is None:
            if BRIGHTFIELD_TIFF_CONFIG.supports_format(path):
                backend = "tifffile"
            elif BRIGHTFIELD_OPENSLIDE_CONFIG.supports_format(path):
                backend = "openslide"
            elif BRIGHTFIELD_MEMORY_CONFIG.supports_format(path):
                backend = "memory"
            else:
                raise ValueError(
                    f"The provided file format ({path.suffix}) is not "
                    + "currently supported."
                )

        match backend:
            case "tifffile":
                self.backend = BrightfieldTiffBackend(path, mpp, level)
            case "openslide":
                self.backend = BrightfieldOpenslideBackend(path, mpp, level)
            case "memory":
                if level:
                    raise ValueError("'level' is not supported for 'memory' backend.")

                self.backend = BrightfieldMemoryBackend(path, mpp)
            case _:
                raise ValueError(
                    "If 'backend' is provided, must be one of: "
                    + "'tifffile', 'openslide', 'memory'."
                )

    def __repr__(self):
        if isinstance((mpp := self.mpp), float):
            mpp = round(self.mpp, 4)

        return f"Brightfield(l={self.l}, h={self.h}, w={self.w}, c={self.c}, mpp={mpp})"

    def __getstate__(self):
        """Handle state for multiprocess pickling."""
        return {"path": self.backend.path}

    def __setstate__(self, state):
        """Recover state for multiprocess pickling."""
        self.__init__(state["path"])

    @property
    def levels(self):
        """Store number of avaiable levels."""
        return self.backend.levels

    @property
    def levels_keys(self):
        """Store keys for each available level."""
        return self.backend.levels_keys

    @property
    def levels_shape(self):
        """Store shape of each available level."""
        return self.backend.levels_shape

    @property
    def l(self):
        """Set current slide level."""
        return self.backend.l

    @property
    def h(self):
        """Set image height."""
        return self.backend.h

    @property
    def w(self):
        """Set image height."""
        return self.backend.w

    @property
    def c(self):
        """Set image channels."""
        return self.backend.c

    @property
    def shape(self):
        """Set image shape"""
        return (self.h, self.w, self.c)

    @property
    def mpp_x(self):
        """Set microns per pixel over x dimension."""
        return self.backend.mpp_x

    @property
    def mpp_y(self):
        """Set microns per pixel over x dimension."""
        return self.backend.mpp_x

    @property
    def mpp(self):
        """Set average microns per pixel."""
        return self.backend.mpp

    def __getitem__(self, key: IndexKey) -> Numeric | NDArray[Numeric]:
        """Direct indexing of the underlying pixel data.

        Parameters
        ----------
        key : IndexKey
            A slice, ellipsis, or any valid index into numpy ndarray.

        Returns
        -------
        Numeric | NDArray[Numeric]
            Subregion of brightfield image.
        """
        return self.backend[key]

    def tile(self, x: int, y: int, w: int, h: int) -> RGBTile:
        """Get an RGB tile specifying an image subregion.

        Parameters
        ----------
        x : int
            Initial x-coordinate of tile.
        y : int
            Initial y-coordinate of tile.
        w : int
            Width of tile
        h : int
            Height of tile.

        Returns
        -------
        RGBTile
            An RGB tile of image subregion.
        """
        if x > self.w or x < 0:
            raise ValueError("x must be greater than 0 and smaller than image width.")

        if y > self.h or y < 0:
            raise ValueError("y must be greater than 0 and smaller than image height.")

        if h == 1 or w < 1:
            raise ValueError("h and w must be greater than 0.")

        return RGBTile(self.backend[y : y + h, x : x + w])

    def thumbnail(
        self,
        t: int = 5000,
        force: bool = False,
        height: int = None,
        width: int = None,
        resample: str | None = "bilinear",
    ) -> RGBTile:
        """Get a lower resolution thumbnail.

        Parameters
        ----------
        t: int
            If pyramidal, find level with height + width closest to this value.
        force: bool
            If True and image is not pyramidal, then the thumbnail will be
            generated at the height and width of the full resolution image.
        height: None | int
            Optionally resize the thumbnail to a user-provided height.
        width: None | int
            Optionally resize the thumbnail to a user-provided width.
        resample: str | None
            Resampling filter ('bilinear', 'bicubic', 'lanczos', 'nearest').
        """
        if force and not self.levels_shape:
            tile = RGBTile(self[:])
        else:
            if not self.levels_shape and (self.w > t or self.h > t):
                return ValueError(
                    "Image only has a single level and current width and "
                    + "height are greater than 't'. This suggests the lowest "
                    + "level may be large. If you still want to generate the "
                    + "thumbnail, re-run `.thumbnail()` with `force=True`."
                )

            closest_key = None
            min_difference = float("inf")

            for key, (h, w, _) in self.levels_shape.items():
                difference = abs(h - t) + abs(w - t)
                if difference < min_difference:
                    min_difference = difference
                    closest_key = key

            tile = RGBTile(Brightfield(self.backend.path, level=closest_key)[:])

        if height or width:
            tile = tile.resize(height, width, resample)

        return tile



import numpy as np
from bamboo.tile import Tiles
from bamboo.image import Brightfield
from bamboo.models.zoo import GrandQC, GrandQCStats
from bamboo.models._semantic import Semantic, SemanticStats
from bamboo.writer import NumpyWriter, SharedMemoryWriter, MemmapWriter, ZarrWriter
from tests._fixtures.images import RgbBmp


def test_grandqc_semantic():
    """Test GrandQC semantic inference with multiple writer types and parallelization."""
    settings = [
        (NumpyWriter, None, 1),
        (SharedMemoryWriter, "processes", 2),
        (SharedMemoryWriter, "threads", 2),
        (MemmapWriter, "processes", 2),
        (MemmapWriter, "threads", 2),
        (ZarrWriter, "processes", 2),
        (ZarrWriter, "threads", 2),
    ]

    model = GrandQC(device="cpu")
    for writer_cls, prefer, jobs in settings:
        image = Brightfield(RgbBmp.path, mpp=model.mpp)

        if writer_cls == MemmapWriter:
            writer = writer_cls(dtype=model.dtype, path="temp.npy")
        elif writer_cls == ZarrWriter:
            writer = writer_cls(dtype=model.dtype, path="temp.zarr")
        else:
            writer = writer_cls(dtype=model.dtype)

        engine = Semantic(
            model,
            Tiles(image, model.tile_width, model.tile_height, mpp=model.mpp),
            writer=writer,
            n_jobs=jobs,
            prefer=prefer,
        )

        msg = f"Failed: {writer.__class__.__name__}, {prefer}"
        output = engine.run(silent=True)
        assert output.shape == image.shape[:2], msg
        assert np.sum(output) > 1, msg

        writer.cleanup()
        del engine


def test_grandqc_semantic_stats():
    """Test GrandQC semantic stats extraction with SemanticStats engine."""
    model = GrandQC(device="cpu")
    image = Brightfield(RgbBmp.path, mpp=model.mpp)

    tiles = Tiles(image, model.tile_width, model.tile_height, mpp=model.mpp)

    engine = SemanticStats(
        model,
        GrandQCStats,
        tiles,
        n_jobs=1,
        prefer=None,
    )

    results = engine.run(silent=True)
    assert len(results) == tiles.n_tiles
    assert "stats" in results[0]

    for category in [
        "TISSUE",
        "FOLDS",
        "SPOTS",
        "MARKINGS",
        "ARTIFACTS",
        "OUT_OF_FOCUS",
        "BACKGROUND",
        "ARTIFACT",
    ]:
        assert category in results[0]["stats"]

    assert "coords" in results[0]
    assert len(results[0]["coords"]) == 6

    del engine



from dataclasses import dataclass

DATA: str = "src/tests/data/"


@dataclass(frozen=True)
class SlideBif:
    path: str = DATA + "bif/OS-1.bif"
    mpp_x: float = 0.23250000000000001
    mpp_y: float = 0.23250000000000001
    width: int = 105813
    height: int = 93951
    channels: int = 3


@dataclass(frozen=True)
class SlideNdpi:
    path: str = DATA + "ndpi/CMU-1.ndpi"
    mpp_x: float = 0.45641259698767683
    mpp_y: float = 0.45506257110352671
    width: int = 51200
    height: int = 38144
    channels: int = 3


@dataclass(frozen=True)
class SlidePhilips:
    path: str = DATA + "philips/Philips-1.tiff"
    mpp_x: float = 0.226907
    mpp_y: float = 0.22689100000000001
    width: int = 45056
    height: int = 35840
    channels: int = 3


@dataclass(frozen=True)
class SlideLeica:
    path: str = DATA + "scn/Leica-1.scn"
    mpp_x: float = 0.5
    mpp_y: float = 0.5
    width: int = 36832
    height: int = 38432
    channels: int = 3


@dataclass(frozen=True)
class SlideSvs:
    path: str = DATA + "svs/CMU-1.svs"
    mpp_x: float = 0.499
    mpp_y: float = 0.499
    width: int = 46000
    height: int = 32914
    channels: int = 3


@dataclass(frozen=True)
class SlideTiff:
    path: str = DATA + "tiff/CMU-1.tiff"
    mpp_x: None | float = None
    mpp_y: None | float = None
    width: int = 46000
    height: int = 32914
    channels: int = 3


@dataclass(frozen=True)
class SlideTrestle:
    path: str = DATA + "trestle/CMU-1.tif"
    mpp_x: float = 0.57469189167022705
    mpp_y: float = 0.57506245374679565
    width: int = 40000
    height: int = 27712
    channels: int = 3


@dataclass(frozen=True)
class SlideOmeTiff:
    path: str = DATA + "xenium/image.ome.tif"
    mpp_x: float = 0.2737653984781153
    mpp_y: float = 0.2737746952056361
    width: int = 14896
    height: int = 27502
    channels: int = 3


@dataclass(frozen=True)
class SlideError:
    path: str = DATA + "README.md"
    mpp_x: float = None
    mpp_y: float = None
    width: int = 0
    height: int = 0
    channels: int = 0



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

from anci import base


@base("brightfield")
def brightfield():
    """Processing and analysis of gigapixel brightfield images."""
    pass


from .nn import brightfield_nn_semantic
from .tile import brightfield_tile_extract

__all__ = ["brightfield_semantic"]



import pytest
import numpy as np

from bamboo.error import ShapeError
from bamboo.image import Brightfield
from bamboo.tile import GenericTile, RGBTile, Tiles
from tests._fixtures.slides import SlideSvs


def test_generic_tile_init():
    tile = GenericTile(np.random.rand(10, 10))
    assert hasattr(tile, "buffer")
    assert hasattr(tile, "dtype")
    assert hasattr(tile, "shape")
    assert hasattr(tile, "h")
    assert hasattr(tile, "w")
    assert hasattr(tile, "c")

    assert tile.shape == (10, 10, 1)
    assert tile.dtype == np.float64
    assert tile.h == 10
    assert tile.w == 10
    assert tile.c == 1


def test_generic_tile_buffer_type_error():
    with pytest.raises(TypeError):
        _ = GenericTile([[10, 10]])


def test_generic_tile_buffer_element_type_error():
    with pytest.raises(TypeError):
        _ = GenericTile(np.random.rand(10, 10).astype(str))


def test_generic_tile_dim_error():
    with pytest.raises(ShapeError):
        _ = GenericTile(np.random.rand(10, 10, 10, 10))


def test_generic_tile_indexing():
    tile = GenericTile(np.random.randint(0, 255, size=(10, 10, 3), dtype=np.uint8))
    h, w, c = tile.h, tile.w, tile.c

    # [int, int, int] to scalar
    assert isinstance(tile[0, 0, 0], (int, np.integer))
    # [slice, int, int] to array
    assert tile[0:10, 0, 0].shape == (10,)
    # [int, slice, int]
    assert tile[0, 0:10, 0].shape == (10,)
    # [int, int, slice]
    assert tile[0, 0, 0:c].shape == (c,)
    # [slice, slice, int]
    assert tile[0:5, 0:5, 0].shape == (5, 5)
    # [slice, int, slice]
    assert tile[0:5, 0, 0:c].shape == (5, c)
    # [int, slice, slice]
    assert tile[0, 0:5, 0:c].shape == (5, c)
    # [slice, slice, slice]
    assert tile[0:2, 0:2, 0:c].shape == (2, 2, c)
    # [...]
    assert tile[...].shape == (h, w, c)
    # [..., int]
    assert tile[..., 0].shape == (h, w)
    # [int, ..., int]
    assert tile[0, ..., 0].shape == (w,)
    # [int]
    assert tile[0].shape == (w, c)
    # [slice]
    assert tile[0:10].shape[0] == 10


def test_generic_tile_min_max():
    tile = GenericTile(np.array([[0, 1, 2], [3, 4, 5]]))
    assert tile.min() == 0
    assert np.all(tile.row_min() == np.array([0, 3]))
    assert np.all(tile.col_min() == np.array([0, 1, 2]))
    assert np.all(tile.channel_min() == np.array([0]))
    assert tile.max() == 5
    assert np.all(tile.row_max() == np.array([2, 5]))
    assert np.all(tile.col_max() == np.array([3, 4, 5]))
    assert np.all(tile.channel_max() == np.array([5]))


def test_rgb_tile_init():
    tile = RGBTile(np.random.randint(0, 255, (10, 10, 3), dtype=np.uint8))
    assert tile.shape == (10, 10, 3)
    assert tile.dtype == np.uint8
    assert tile.h == 10
    assert tile.w == 10
    assert tile.c == 3


def test_rgb_tile_type_error():
    with pytest.raises(TypeError):
        _ = RGBTile(np.random.rand(10, 10, 3))


def test_rgb_tile_dim_error():
    with pytest.raises(ShapeError):
        _ = RGBTile(np.random.rand(10, 10).astype(np.uint8))


def test_rgb_tile_channel_error():
    with pytest.raises(ShapeError):
        _ = RGBTile(np.random.randint(0, 255, (10, 10, 4), dtype=np.uint8))


def test_rgb_tile_resize():
    tile = RGBTile(np.random.randint(0, 255, (10, 10, 3), dtype=np.uint8))
    resized = tile.resize(5, 10)
    assert resized.shape == (10, 5, 3)


def test_tiles_basic_iteration():
    image = Brightfield(SlideSvs.path, level="2")
    for tile in Tiles(image, 128, 128):
        assert tile.shape == (128, 128, 3)


def test_tiles_overlap():
    image = Brightfield(SlideSvs.path, level="2")
    tiles = Tiles(image, tile_w=128, tile_h=128, overlap_x=64, overlap_y=64)

    tile_one = next(tiles)
    tile_two = next(tiles)

    # Check that overlapping region matches
    assert np.all(tile_one[:, 64:] == tile_two[:, :64])


class DummyImage:
    """Mock image with minimal API to support Tiles."""

    def __init__(self, w, h):
        self.w = w
        self.h = h
        self.mpp = 0.5  # microns per pixel

    def tile(self, x, y, w, h):
        # Return a dummy placeholder to simulate a Tile object
        return f"tile@({x},{y},{w},{h})"

    def resize(self, w, h, resample=None):
        return self

    def thumbnail(self, **kwargs):
        return self

    def to_numpy(self):
        import numpy as np

        return np.zeros((10, 10, 3), dtype=np.uint8)

    @property
    def shape(self):
        return (self.h, self.w, 3)


def test_exclude_overlaps_removes_expected_tiles():
    """Ensure that overlapping tiles are correctly excluded."""
    img = DummyImage(2048, 2048)
    tiles = Tiles(image=img, tile_w=512, tile_h=512)

    # Expect a 4x4 grid of tiles for a 2048x2048 image
    assert len(tiles) == 16

    # Exclude region overlapping roughly the top-left 2x2 tiles
    exclude = [(0, 0, 600, 600)]

    # Apply exclusion
    tiles.exclude(exclude)

    # Should remove 4 tiles: (0,0), (512,0), (0,512), (512,512)
    remaining = len(tiles)
    assert remaining == 12

    # Ensure no remaining tiles overlap with the exclusion area
    for coord in tiles.coordinates:
        x, y, w, h, *_ = coord
        # Simple overlap check
        overlap = x < 600 and x + w > 0 and y < 600 and y + h > 0
        assert not overlap, f"Tile at ({x},{y}) should have been excluded"


def test_exclude_overlaps_no_overlap_keeps_all():
    """Tiles with no overlap should remain unaffected."""
    img = DummyImage(1024, 1024)
    tiles = Tiles(image=img, tile_w=256, tile_h=256)
    n_before = len(tiles)

    # Exclude region far outside the image
    exclude = [(2000, 2000, 256, 256)]
    tiles.exclude(exclude)

    # Should keep all tiles
    assert len(tiles) == n_before



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

import os

CACHE_KEY: str = "BAMBOO_CACHE"

if CACHE_KEY not in os.environ:
    os.environ[CACHE_KEY] = os.path.expanduser(f"~/.{CACHE_KEY.lower()}")

CACHE_DIR = os.environ[CACHE_KEY]
os.makedirs(CACHE_DIR, exist_ok=True)

from ._semantic import Semantic, SemanticStats

__all__ = ["Semantic", "SemanticStats"]



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

# TO-DO: Resize functionality
# if resize:
#     with tempfile.NamedTemporaryFile(
#         dir=output.parent, suffix=output.suffix, delete=False
#     ) as tmp:
#         tmp_path = Path(tmp.name)

#     try:
#         match output.suffix:
#             case ".npy":
#                 with tempfile.NamedTemporaryFile(
#                     dir=output.parent, suffix=".npy", delete=False
#                 ) as tmp:
#                     tmp_path = Path(tmp.name)

#                 resize_mask_memmap(
#                     input_path=str(output),
#                     output_path=str(tmp_path),
#                     target_shape=(image.h, image.w),
#                     row_block=model.tile_width,
#                     col_block=model.tile_height,
#                 )

#                 os.replace(tmp_path, output)

#             case ".zarr":
#                 tmp_dir = tempfile.TemporaryDirectory(dir=output.parent)
#                 tmp_path = Path(tmp_dir.name) / output.name

#                 resize_mask_zarr(
#                     input_path=str(output),
#                     output_path=str(tmp_path),
#                     target_shape=(image.h, image.w),
#                     row_block=model.tile_width,
#                     col_block=model.tile_height,
#                 )

#                 if output.exists():
#                     shutil.rmtree(output)
#                 os.replace(tmp_path, output)

#     except Exception as e:
#         if tmp_path.exists():
#             try:
#                 tmp_path.unlink()
#             except Exception as cleanup_error:
#                 print(f"Failed to clean-up temporary file: {cleanup_error}")
#         raise e



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

from __future__ import annotations

import numpy as np
import matplotlib.pyplot as plt

from dataclasses import dataclass
from pathlib import Path
from PIL import Image

from bamboo.error import ShapeError
from bamboo.tile._base import Tile

RGBTILE_PIL_FORMATS: set[str] = {
    "avif",
    "blp",
    "bmp",
    "dds",
    "dib",
    "eps",
    "gif",
    "icns",
    "ico",
    "im",
    "jpeg",
    "mpo",
    "msp",
    "pcx",
    "pfm",
    "png",
    "ppm",
    "qoi",
    "sgi",
    "spider",
    "tga",
    "tiff",
    "webp",
    "xbm",
}

RGBTILE_WRITE_FORMATS: set[str] = RGBTILE_PIL_FORMATS | {"npy"}


@dataclass(frozen=True)
class RGBTile(Tile):
    """In-memory two-dimensional RGB tile defining an image subregion.

    Attributes
    ----------
    buffer : np.ndarray
        Two-dimensional 8-bit RGB pixel container (channel in last index).
    dtype : np.dtype
        Data type of buffer elements (subpixels).
    shape : Tuple[int, int, int]
        Height, width, and channels for the tile.
    h : int
        Height of the tile.
    w : int
        Width of the tile.
    c : int
        Number of channels in the tile.
    """

    def __post_init__(self):
        """Buffer validation and attribute initialization."""
        if not isinstance(self.buffer, np.ndarray):
            raise TypeError("[RGBTile] 'buffer' must be of type np.ndarray.")

        dtype = self.buffer.dtype
        if not np.isdtype(dtype, np.uint8):
            raise TypeError("[RGBTile] 'buffer' must be uint8 type.")

        object.__setattr__(self, "dtype", dtype)

        if self.buffer.ndim != 3:
            raise ShapeError("[RGBTile] 'buffer' must be have 3 dimensions.")

        if self.buffer.shape[-1] != 3:
            raise ShapeError("[RGBTile] 'buffer' must have 3 channels.")

        shape = self.buffer.shape
        object.__setattr__(self, "shape", shape)
        object.__setattr__(self, "h", shape[0])
        object.__setattr__(self, "w", shape[1])
        object.__setattr__(self, "c", 3)

    def show(
        self,
        ax: None | plt.Axes = None,
        figsize: None | tuple[int, int] = None,
        hide_axes: bool = True,
    ) -> plt.Axes:
        """Visualize the tile.

        Parameters
        ----------
        ax : None | plt.Axes
            Optional matplot axes object.
        figsize : None | tuple[int, int]
            Option figure size.
        hide_axes : bool
            If True, then all x-axis, y-axis, and spines are removed.

        Returns
        -------
        plt.Axes
            Matplotlib axes object of tile.
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        ax.imshow(self.buffer)

        if hide_axes:
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)
            for pos in ["left", "right", "top", "bottom"]:
                ax.spines[pos].set_visible(False)

        return ax

    def resize(
        self,
        w: None | int = None,
        h: None | int = None,
        resample: str | None = "bilinear",
    ) -> RGBTile:
        """Resize the tile.

        Parameters
        ----------
        w : None | int
            Resize width.
        h : None | int
            Resize height.
        resample : str | None
            Resampling filter ('bilinear', 'bicubic', 'lanczos', 'nearest')
        """
        match resample:
            case "bilinear":
                resample = Image.Resampling.BILINEAR
            case "bicubic":
                resample = Image.Resampling.BICUBIC
            case "lanczos":
                resample = Image.Resampling.LANCZOS
            case "nearest":
                resample = Image.Resampling.NEAREST
            case _:
                raise ValueError(
                    "'resample' must be one of: "
                    + "'bilinear', 'bicubic', 'lanczos', 'nearest'."
                )

        if h is None and w is None:
            raise ValueError("One of 'height' or 'width' must be specified.")

        if h is None:
            h = int(self.h * w / self.w)

        if w is None:
            w = int(self.w * h / self.h)

        # Note: PIL Image resize is not particularly fast - so it may be useful
        # to write an interface to the fast_image_resize rust crate
        return RGBTile(np.array(Image.fromarray(self.buffer).resize((w, h), resample)))

    def write(
        self,
        path: str | Path,
    ) -> None:
        """Write the RGBTile to disk.

        Parameters
        ----------
        path : str | Path
            A path specifying file with valid image extension.

        Returns
        -------
        None
            RGBTile is written to disk.
        """
        if isinstance(path, str):
            path = Path(path)

        if path.suffix[1:] in RGBTILE_PIL_FORMATS:
            Image.fromarray(self.buffer).save(path)
        elif path.suffix == ".npy":
            np.save(path, self.buffer)
        else:
            raise ValueError(
                "'RGBTile' write failed. 'path' must specify a supported image format: "
                f"{RGBTILE_WRITE_FORMATS}"
            )



import pytest
import numpy as np
from pathlib import Path
from bamboo.image import Brightfield
from bamboo.tile import RGBTile

from tests._fixtures.slides import (
    SlideBif,
    SlideNdpi,
    SlidePhilips,
    SlideLeica,
    SlideSvs,
    SlideTiff,
    SlideTrestle,
    SlideOmeTiff,
    SlideError,
)

from tests._fixtures.images import (
    RgbBmp,
    RgbJpeg,
    RgbNpy,
    RgbPbm,
    RgbPng,
    RgbTga,
    RgbWebp,
)


def test_init_str():
    image = Brightfield(SlideSvs.path)
    for attr in [
        "backend",
        "levels",
        "levels_keys",
        "levels_shape",
        "l",
        "h",
        "w",
        "c",
        "mpp",
        "mpp_x",
        "mpp_y",
    ]:
        assert hasattr(image, attr)


def test_init_pathlib():
    image = Brightfield(Path(SlideSvs.path))
    for attr in [
        "backend",
        "levels",
        "levels_keys",
        "levels_shape",
        "l",
        "h",
        "w",
        "c",
        "mpp",
        "mpp_x",
        "mpp_y",
    ]:
        assert hasattr(image, attr)


def test_init_error_path():
    with pytest.raises(ValueError):
        _ = Brightfield(SlideError.path)


def test_slides_auto():
    for slide in [
        SlideBif,
        SlideNdpi,
        SlidePhilips,
        SlideLeica,
        SlideSvs,
        SlideTiff,
        SlideOmeTiff,
    ]:
        _ = Brightfield(slide.path)

    with pytest.raises(ValueError):
        _ = Brightfield(SlideTrestle.path)


def test_images_auto():
    for image in [RgbBmp, RgbJpeg, RgbNpy, RgbPbm, RgbPng, RgbTga, RgbWebp]:
        _ = Brightfield(image.path)


def test_backend_tifffile():
    _ = Brightfield(SlideSvs.path, backend="tifffile")


def test_backend_openslide():
    _ = Brightfield(SlideSvs.path, backend="openslide")


def test_backend_memory():
    _ = Brightfield(RgbBmp.path, backend="memory")


def test_indexing():
    image = Brightfield(SlideSvs.path, level="2")
    h, w, c = image.backend.h, image.backend.w, 3

    assert isinstance(image[0, 0, 0], (int, np.integer))
    assert image[0:10, 0, 0].shape == (10,)
    assert image[0, 0:10, 0].shape == (10,)
    assert image[0, 0, 0:c].shape == (c,)
    assert image[0:5, 0:5, 0].shape == (5, 5)
    assert image[0:5, 0, 0:c].shape == (5, c)
    assert image[0, 0:5, 0:c].shape == (5, c)
    assert image[0:2, 0:2, 0:c].shape == (2, 2, c)
    assert image[...].shape == (h, w, c)
    assert image[..., 0].shape == (h, w)
    assert image[0, ..., 0].shape == (w,)
    assert image[0].shape == (w, c)
    assert image[0:10].shape[0] == 10


def test_tile():
    image = Brightfield(SlideSvs.path)
    tile = image.tile(0, 0, 512, 256)
    assert tile.w == 512
    assert tile.h == 256


def test_thumbnail_default():
    image = Brightfield(SlideSvs.path)
    h, w, c = list(image.levels_shape.values())[-1]
    thumbnail = image.thumbnail(t=0)
    assert isinstance(thumbnail, RGBTile)
    assert thumbnail.h == h
    assert thumbnail.w == w
    assert thumbnail.c == c


def test_thumbnail_resize():
    image = Brightfield(SlideSvs.path)
    thumbnail = image.thumbnail(height=100, width=100)
    assert thumbnail.h == 100
    assert thumbnail.w == 100
    assert thumbnail.c == 3


def test_thumbnail_resize_height():
    image = Brightfield(SlideSvs.path)
    h, w = image.h, image.w
    thumbnail = image.thumbnail(height=1000)
    tolerance = 1 / min(thumbnail.w, thumbnail.h)
    assert abs(thumbnail.w / thumbnail.h - w / h) < tolerance


def test_thumbnail_resize_width():
    image = Brightfield(SlideSvs.path)
    h, w = image.h, image.w
    thumbnail = image.thumbnail(width=1000)
    tolerance = 1 / min(thumbnail.w, thumbnail.h)
    assert abs(thumbnail.w / thumbnail.h - w / h) < tolerance



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

import numpy as np
import torch

from dataclasses import dataclass
from numpy.typing import NDArray
from typing import Tuple

from bamboo.types import IndexKey, Numeric


@dataclass(frozen=True)
class Tile:
    """A base class for a two-dimensional in-memory tile of image subregion.

    Attributes
    ----------
    buffer : np.ndarray
        Two-dimensional pixel container with optional last index channel.
    dtype : np.dtype
        Data type of buffer elements (subpixels).
    shape : Tuple[int, int, int]
        Height, width, and channels for the tile.
    h : int
        Height of the tile.
    w : int
        Width of the tile.
    c : int
        Number of channels in the tile.
    """

    buffer: np.ndarray
    dtype: Numeric = None
    shape: Tuple[int, int, int] = None
    h: int = None
    w: int = None
    c: int = None

    def __getitem__(self, key: IndexKey) -> Numeric | NDArray[Numeric]:
        """Direct indexing of the underlying buffer.

        Parameters
        ----------
        key : IndexKey
            A slice, ellipsis, or any valid index into numpy ndarray.

        Returns
        -------
        Numeric | NDArray[Numeric]
            Slice
        """
        return self.buffer[key]

    def to_numpy(self) -> NDArray[Numeric]:
        """Return tile as a numpy array.

        Returns
        -------
        NDArray[Numeric]
            Pixels in numpy array format.
        """
        return self.buffer[:]

    def to_tensor(self) -> torch.Tensor:
        """Return tile as a torch tensor.

        Returns
        -------
        torch.Tensor
            Pixels in torch tensor format.
        """
        return torch.from_numpy(self.buffer[:])

    def min(self) -> Numeric:
        """Get minimum subpixel value in tile.

        Returns
        -------
        Numeric
            Minimum subpixel value.
        """
        return self.buffer.min()

    def max(self) -> Numeric:
        """Get maximum subpixel value in tile.

        Returns
        -------
        Numeric
            Maximum subpixel value.
        """
        return self.buffer.max()

    def row_min(self) -> NDArray[Numeric]:
        """Get minimum subpixel values in each row of tile.

        Returns
        -------
        NDArray[Numeric]
            Minimum subpixel values per row.
        """
        return self.buffer.min(axis=(1, 2))

    def row_max(self) -> NDArray[Numeric]:
        """Get maximum subpixel values in each row of tile.

        Returns
        -------
        NDArray[Numeric]
            Maximum subpixel values per row.
        """
        return self.buffer.max(axis=(1, 2))

    def col_min(self) -> NDArray[Numeric]:
        """Get minimum subpixel values in each column of tile.

        Returns
        -------
        NDArray[Numeric]
            Minimum subpixel values per column.
        """
        return self.buffer.min(axis=(0, 2))

    def col_max(self) -> NDArray[Numeric]:
        """Get maximum subpixel values in each column of tile.

        Returns
        -------
        NDArray[Numeric]
            Maximum subpixel values per column.
        """
        return self.buffer.max(axis=(0, 2))

    def channel_min(self) -> NDArray[Numeric]:
        """Get minimum subpixel values in each channel of tile.

        Returns
        -------
        NDArray[Numeric]
            Minimum subpixel values per channel.
        """
        return self.buffer.min(axis=(0, 1))

    def channel_max(self) -> NDArray[Numeric]:
        """Get maximum subpixel values in each channel of tile.

        Returns
        -------
        NDArray[Numeric]
            Maximum subpixel values per channel.
        """
        return self.buffer.max(axis=(0, 1))



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

import numpy as np
import zarr

from typing import Tuple, Any
from types import EllipsisType

# A general numeric type for integer and floating point types
Numeric = int | float | np.integer[Any] | np.floating[Any]

# Type alias for a single element of an indexing key
IndexElement = int | slice | EllipsisType | None

# Type alias for single indices, slices, Ellipsis, and associated tuples.
IndexKey = IndexElement | Tuple[IndexElement, ...]

# Valid output array types
Array = np.ndarray | np.memmap | zarr.Array



# Copyright (c) 2025, Tom Ouellette
# Licensed under the GNU GPLv3 License

import os
import numpy as np
import shutil
import zarr

from abc import abstractmethod, ABC
from functools import wraps
from multiprocessing import shared_memory
from pathlib import Path
from typing import Callable


class ArrayWriter(ABC):
    """A base class for array writers.

    Any array writer may have shape and data type set at initialization or
    set dynamically after initialization. The post-initialization setters
    are used in cases where the shape and data type of the output array is
    hidden from the user. An optional context manager is also available.

    Parameters
    ----------
    shape : None | tuple[int, ...]
        Output array shape e.g. (h, w, c).
    dtype : None | np.dtype
        Output array data type e.g. np.uint8.

    Attributes
    ----------
    shape : tuple of int
        Output array shape.
    dtype : data-type
        Output array data type.
    output : nd.ndarray or None
        Created array or `None` if `create` has not been called yet.
    """

    def __init__(
        self, shape: None | tuple[int, ...] = None, dtype: None | np.dtype = None
    ):
        self._shape = shape
        self._dtype = dtype
        self.output = None

    @property
    def shape(self) -> tuple[int, ...]:
        return self._shape

    @shape.setter
    def shape(self, shape: tuple[int, ...]):
        if not isinstance(shape, tuple) or not all(isinstance(s, int) for s in shape):
            raise TypeError("'shape' must be a tuple of integers.")
        self._shape = shape

    @property
    def dtype(self) -> np.dtype:
        return self._dtype

    @dtype.setter
    def dtype(self, dtype: np.dtype):
        self._dtype = np.dtype(dtype)

    @abstractmethod
    def create(self):
        """Allocate the array."""
        pass

    @abstractmethod
    def cleanup(self):
        """Release resources."""
        pass

    def __enter__(self):
        if self.output is None:
            self.create()
        return self.output

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()


class NumpyWriter(ArrayWriter):
    """A basic in-memory numpy array writer."""

    def create(self):
        """
        Create a zero-initialized NumPy array.

        Returns
        -------
        ndarray
            A NumPy array initialized with zeros.
        """
        if self._shape is None or self._dtype is None:
            raise ValueError("'shape' and 'dtype' must be set before array creation.")

        self.output = np.zeros(self._shape, dtype=self._dtype)

    def cleanup(self):
        """No cleanup required for in-memory numpy arrays."""
        self.output = None


class SharedMemoryWriter(ArrayWriter):
    """An array writer backed by shared memory.

    Attributes
    ----------
    shm : shared_memory.SharedMemory
        Shared memory block used to store the array.
    """

    def __init__(
        self, shape: None | tuple[int, ...] = None, dtype: None | np.dtype = None
    ):
        super().__init__(shape, dtype)
        self.shm: shared_memory.SharedMemory = None

    def create(self):
        """Creates a shared memory-backed array."""
        if self._shape is None or self._dtype is None:
            raise ValueError("'shape' and 'dtype' must be set before array creation.")

        size = np.prod(self._shape) * np.dtype(self._dtype).itemsize
        self.shm = shared_memory.SharedMemory(create=True, size=size)
        self.output = np.ndarray(self._shape, dtype=self._dtype, buffer=self.shm.buf)
        self.output.setflags(write=True)

    def cleanup(self):
        """Clean up shared memory resources safely."""
        if self.output is not None:
            self.output = None
        if self.shm is not None:
            try:
                self.shm.close()
            except Exception:
                pass
            try:
                self.shm.unlink()
            except Exception:
                pass
            finally:
                self.shm = None


class MemmapWriter(ArrayWriter):
    """A memory-mapped array writer.

    Parameters
    ----------
    shape : tuple[int, ...]
        Array shape.
    dtype : np.dtype
        Array data type.
    path : str | Path
        File path where the memory-mapped file will be stored.

    Attributes
    ----------
    path : str
        File path of the memory-mapped file.
    """

    def __init__(
        self,
        path: str | Path,
        shape: None | tuple[int, ...] = None,
        dtype: None | np.dtype = None,
    ):
        super().__init__(shape, dtype)
        if not str(path).endswith(".npy"):
            raise ValueError("'path' must be specify a .npy file.")

        self.path = path

    def create(self):
        """Create a memory-mapped array on disk."""
        if self._shape is None or self._dtype is None:
            raise ValueError("'shape' and 'dtype' must be set before array creation.")

        self.output = np.lib.format.open_memmap(
            self.path, dtype=self._dtype, shape=self._shape, mode="w+"
        )

    def cleanup(self):
        """Clean up memory-mapped file resources by deleting the file."""
        if self.output is not None:
            if hasattr(self.output, "flush"):
                self.output.flush()
            del self.output
            self.output = None
        if os.path.exists(self.path):
            try:
                os.remove(self.path)
            except OSError:
                pass


class ZarrWriter(ArrayWriter):
    """A zarr array writer.

    Parameters
    ----------
    path : str
        File path where the .zarr file will be stored.
    chunks : tuple[int, ...]
        Chunks shape for the zarr array.
    shape : tuple[int, ...]
        Array shape.
    dtype : np.dtype
        Array data type.

    Attributes
    ----------
    path : str | Path
        File path of the zarr array.
    chunks : tuple[int, ...]
        Chunk shape.
    output : zarr.Array | None
        Initialized Zarr array once `create` is called.
    """

    def __init__(
        self,
        path: str | Path,
        chunks: None | tuple[int, ...] = None,
        shape: None | tuple[int, ...] = None,
        dtype: None | np.dtype = None,
    ):
        super().__init__(shape, dtype)
        if not str(path).endswith(".zarr"):
            raise ValueError("'path' must be specify a .zarr file.")

        self.path = path
        self.chunks = chunks

    def create(self):
        """Create a zarr array on disk."""
        if self._shape is None or self._dtype is None or self.chunks is None:
            raise ValueError(
                "'shape', 'dtype', and 'chunks' must be set before zarr array creation."
            )

        self.output = zarr.open(
            store=self.path,
            mode="w",
            shape=self._shape,
            chunks=self.chunks,
            dtype=self._dtype,
        )

    def cleanup(self):
        """Clean up .zarr file by deleting the file."""
        if os.path.exists(self.path):
            shutil.rmtree(self.path)


def inject_writer(writer: ArrayWriter) -> Callable:
    """Wraps a worker function to inject writer-specific context.

    This decorator should be used on any worker function that may
    use any of the possible derived ArrayWriter classes. For certain
    writers, additional data must be passed to each worker to ensure
    arrays can be written to in parallel. For any writer, the worker
    function must specify the output array in the first position.

    Parameters
    ----------
    writer : ArrayWriter
        A valid ArrayWriter.

    Returns
    -------
    Callable
        Decorated function with injected writer-specific data.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if isinstance(writer, SharedMemoryWriter):
                shm = shared_memory.SharedMemory(name=writer.shm.name)
                arr = np.ndarray(writer._shape, dtype=writer._dtype, buffer=shm.buf)
                try:
                    return func(arr, *args, **kwargs)
                finally:
                    shm.close()
            elif isinstance(writer, ZarrWriter):
                arr = zarr.open(writer.path, mode="r+")
                return func(arr, *args, **kwargs)
            else:
                return func(writer.output, *args, **kwargs)

        return wrapper

    return decorator


def get_array_writer(
    shape: tuple[int, ...],
    backend: str,
    dtype: np.dtype,
    path: None | str = None,
    chunks: None | tuple[int, ...] = None,
) -> ArrayWriter:
    """Initialize an array writer backend.

    Parameters
    ----------
    shape : tuple[int, ...]
        Shape of array being written to.
    backend : str
        Array writer backend. One of 'numpy', 'shared', 'memmap'.
    path : None | str
        If backend is 'memmap' or 'zarr', then path to output file.
    chunks : tuple[int, ...]
        If backend is 'zarr', then Chunks shape for the zarr array.

    Returns
    -------
    ArrayWriter
        Array writer with specified backend.
    """
    match backend:
        case "numpy":
            return NumpyWriter(shape, dtype)
        case "shared":
            return SharedMemoryWriter(shape, dtype)
        case "memmap":
            if path is None:
                raise ValueError("If backend is 'memmap', 'path' must be specified.")
            return MemmapWriter(path, shape, dtype)
        case "zarr":
            if path is None:
                raise ValueError("If backend is 'zarr', 'path' must be specified.")
            if chunks is None:
                raise ValueError("If backend is 'zarr', 'chunks' must be specified.")
            return ZarrWriter(path, chunks, shape, dtype)
        case _:
            raise ValueError(
                "'backend' must be one of: 'numpy', 'shared', 'memmap', 'zarr'."
            )
